# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import Any

from horizon.commons.exceptions.base import ApplicationError


class EntityNotFoundError(ApplicationError):
    """Entity not found.

    Examples
    --------

    >>> from horizon.commons.exceptions import EntityNotFoundError
    >>> raise EntityNotFoundError("User", "username", "test")
    Traceback (most recent call last):
    horizon.commons.exceptions.entity.EntityNotFoundError: User with username='test' not found
    """

    entity_type: str
    """Entity type"""

    field: str
    """Entity identifier field"""

    value: Any
    """Entity identifier value"""

    def __init__(self, entity_type: str, field: str, value: Any):
        self.entity_type = entity_type
        self.field = field
        self.value = value

    @property
    def message(self) -> str:
        if self.field is not None:
            return f"{self.entity_type} with {self.field}={self.value!r} not found"
        return f"{self.entity_type} not found"

    @property
    def details(self) -> dict[str, Any]:
        return {
            "entity_type": self.entity_type,
            "field": self.field,
            "value": self.value,
        }


class EntityAlreadyExistsError(ApplicationError):
    """Entity with same identifier already exists.

    Examples
    --------


    >>> from horizon.commons.exceptions import EntityNotFoundError
    >>> raise EntityAlreadyExistsError("User", "username", "test")
    Traceback (most recent call last):
    horizon.commons.exceptions.entity.EntityAlreadyExistsError: User with username='test' already exists
    """

    entity_type: str
    """Entity type"""

    field: str
    """Entity identifier field"""

    value: Any
    """Entity identifier value"""

    def __init__(self, entity_type: str, field: str, value: Any):
        self.entity_type = entity_type
        self.field = field
        self.value = value

    @property
    def message(self) -> str:
        return f"{self.entity_type} with {self.field}={self.value!r} already exists"

    @property
    def details(self) -> dict[str, Any]:
        return {
            "entity_type": self.entity_type,
            "field": self.field,
            "value": self.value,
        }



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
import http

from pydantic import BaseModel
from typing_extensions import Literal

from horizon.commons.errors.base import BaseErrorSchema
from horizon.commons.errors.registration import register_error_response
from horizon.commons.exceptions import PermissionDeniedError


class PermissionDeniedDetailsSchema(BaseModel):
    required_role: str
    actual_role: str


@register_error_response(
    exception=PermissionDeniedError,
    status=http.HTTPStatus.FORBIDDEN,
)
class PermissionDeniedSchema(BaseErrorSchema):
    code: Literal["permission_denied"] = "permission_denied"
    details: PermissionDeniedDetailsSchema

    def to_exception(self) -> PermissionDeniedError:
        return PermissionDeniedError(
            required_role=self.details.required_role,
            actual_role=self.details.actual_role,
        )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from horizon.commons.dto.pagination import Pagination
from horizon.commons.dto.unset import Unset

__all__ = [
    "Pagination",
    "Unset",
]



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from horizon.backend.db.models import HWMHistory
from horizon.backend.db.repositories.base import Repository
from horizon.commons.dto import Pagination


class HWMHistoryRepository(Repository[HWMHistory]):
    async def paginate(
        self,
        hwm_id: int,
        page: int,
        page_size: int,
    ) -> Pagination[HWMHistory]:
        return await self._paginate(
            where=[
                HWMHistory.hwm_id == hwm_id,
            ],
            order_by=[HWMHistory.changed_at.desc()],
            page=page,
            page_size=page_size,
        )

    async def create(self, hwm_id: int, data: dict) -> HWMHistory:
        action = data.get("action", "Created")

        result = await self._create(
            data={
                **data,
                "hwm_id": hwm_id,
                "action": action,
            },
        )
        await self._session.flush()
        return result

    async def bulk_create(self, hwm_data: list[dict]) -> list[HWMHistory]:
        hwm_histories = [HWMHistory(**data) for data in hwm_data]
        self._session.add_all(hwm_histories)
        await self._session.flush()
        return hwm_histories



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from pydantic import BaseModel, Field


class ApplicationVersionSettings(BaseModel):
    """X-Application-Version Middleware Settings.

    Examples
    --------

    .. code-block:: bash

        HORIZON__SERVER__APPLICATION_VERSION__ENABLED=True
        HORIZON__SERVER__APPLICATION_VERSION__HEADER_NAME=X-Application-Version
    """

    enabled: bool = Field(default=True, description="Set to ``True`` to enable middleware")
    header_name: str = Field(
        default="X-Application-Version",
        description="Name of response header which is filled up with application version number",
    )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from horizon.client.auth.access_token import AccessToken
from horizon.client.auth.base import BaseAuth
from horizon.client.auth.login_password import LoginPassword

__all__ = [
    "AccessToken",
    "BaseAuth",
    "LoginPassword",
]



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

"""
Settings for LDAPAuthProvider class.

Basic LDAP terminology is explained here: `LDAP Overview <https://www.zytrax.com/books/ldap/ch2/>`_
"""

import textwrap
from typing import Any, Dict

from pydantic import BaseModel, Field

from horizon.backend.settings.auth.ldap import LDAPAuthProviderSettings


class LDAPCachePasswordHashSettings(BaseModel):
    """Settings related to LDAP credentials cache password hashing.

    Examples
    --------

    .. code-block:: bash

        HORIZON__AUTH__CACHE__PASSWORD_HASH__ALGORITHM=argon2
        HORIZON__AUTH__CACHE__PASSWORD_HASH__OPTIONS={"time_cost": 2, "memory_cost": 1024, "parallelism": 1}
    """

    algorithm: str = Field(
        default="argon2",
        description=textwrap.dedent(
            """
            Hashing algorithm used to hash user credentials.

            See `passlib documentation <https://passlib.readthedocs.io/en/stable/lib/passlib.hash.html#active-hashes>`_
            for more details.
            """,
        ),
    )
    options: Dict[str, Any] = Field(
        default={},
        description="Options passed to hashing algorithm",
    )


class LDAPCacheSettings(BaseModel):
    """Settings related to LDAP credentials cache.

    Examples
    --------

    .. code-block:: bash

        HORIZON__AUTH__CACHE__EXPIRE_SECONDS=3600  # 1 hour
    """

    expire_seconds: int = Field(
        default=60 * 60,
        description=textwrap.dedent(
            """
            Credentials cache expiration time, in seconds.

            .. warning::

                Please do not set too large value here, as it may lead to security issues.
            """,
        ),
    )
    password_hash: LDAPCachePasswordHashSettings = Field(
        default_factory=LDAPCachePasswordHashSettings,
        description="Password hashing options",
    )


class CachedLDAPAuthProviderSettings(LDAPAuthProviderSettings):
    """Settings for CachedLDAPAuthProvider.

    Examples
    --------

    .. code-block:: bash

        HORIZON__AUTH__PROVIDER=horizon.backend.providers.auth.cached_ldap.CachedLDAPAuthProvider
        HORIZON__AUTH__ACCESS_KEY__SECRET_KEY=secret
        HORIZON__AUTH__LDAP__URL=ldap://ldap.domain.com:389
        HORIZON__AUTH__LDAP__LOOKUP__ENABLED=True
        HORIZON__AUTH__LDAP__LOOKUP__POOL__ENABLED=True
        HORIZON__AUTH__LDAP__LOOKUP__CREDENTIALS__USER=uid=techuser,ou=users,dc=example,dc=com
        HORIZON__AUTH__LDAP__LOOKUP__CREDENTIALS__PASSWORD=somepassword
        HORIZON__AUTH__CACHE__EXPIRE_SECONDS=3600  # 1 hour
    """

    cache: LDAPCacheSettings = Field(default_factory=LDAPCacheSettings, description="Cache related settings")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from pydantic import BaseModel, Field

from horizon.backend.settings.auth.jwt import JWTSettings


class DummyAuthProviderSettings(BaseModel):
    """Settings for DummyAuthProvider.

    Examples
    --------

    .. code-block:: bash

        HORIZON__AUTH__PROVIDER=horizon.backend.providers.auth.dummy.DummyAuthProvider
        HORIZON__AUTH__ACCESS_KEY__SECRET_KEY=secret
    """

    access_token: JWTSettings = Field(description="Access-token related settings")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
import http

from typing_extensions import Literal

from horizon.commons.errors.base import BaseErrorSchema
from horizon.commons.errors.registration import register_error_response
from horizon.commons.exceptions import BadRequestError


@register_error_response(
    exception=BadRequestError,
    status=http.HTTPStatus.BAD_REQUEST,
)
class BadRequestSchema(BaseErrorSchema):
    code: Literal["bad_request"] = "bad_request"



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from horizon.backend.db.models import NamespaceHistory
from horizon.backend.db.repositories.base import Repository
from horizon.commons.dto import Pagination


class NamespaceHistoryRepository(Repository[NamespaceHistory]):
    async def paginate(
        self,
        namespace_id: int,
        page: int,
        page_size: int,
    ) -> Pagination[NamespaceHistory]:
        return await self._paginate(
            where=[
                NamespaceHistory.namespace_id == namespace_id,
            ],
            order_by=[NamespaceHistory.changed_at.desc()],
            page=page,
            page_size=page_size,
        )

    async def create(self, namespace_id: int, data: dict) -> NamespaceHistory:
        action = data.get("action", "Created")

        result = await self._create(
            data={
                **data,
                "namespace_id": namespace_id,
                "action": action,
            },
        )
        await self._session.flush()
        return result



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from horizon.backend.db.repositories.credentials_cache import CredentialsCacheRepository
from horizon.backend.db.repositories.hwm import HWMRepository
from horizon.backend.db.repositories.hwm_history import HWMHistoryRepository
from horizon.backend.db.repositories.namespace import NamespaceRepository
from horizon.backend.db.repositories.namespace_history import NamespaceHistoryRepository
from horizon.backend.db.repositories.user import UserRepository



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from urllib.parse import urlparse

from pydantic import AnyHttpUrl, BaseModel, SecretStr
from typing_extensions import Literal

from horizon.client.auth.base import BaseAuth


class LoginPassword(BaseAuth, BaseModel):
    """Authorization using OAuth2 + ``grant_type=password``.

    Resulting access is passed in ``Authorization: Bearer ${token}`` header.
    Tokens can be refreshed.

    Parameters
    ----------
    username : str
        User name

    password : str
        User password

    Examples
    --------

    >>> from horizon.client.auth import LoginPassword
    >>> auth = LoginPassword(login="me", password="12345")
    """

    login: str
    password: SecretStr

    type: Literal["login_password"] = "login_password"

    def patch_session(self, session):
        return session

    def fetch_token_kwargs(self, base_url: AnyHttpUrl) -> dict[str, str]:
        # default path for token
        parsed_url = urlparse(str(base_url))
        token_url = parsed_url._replace(path=parsed_url.path + "/v1/auth/token")  # noqa: WPS437, WPS336
        url = token_url.geturl()

        return {
            "url": str(url),
            "username": self.login,
            "password": self.password.get_secret_value(),
        }



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from fastapi import APIRouter, Depends, FastAPI, Request
from prometheus_client import (
    CONTENT_TYPE_LATEST,
    CollectorRegistry,
    Gauge,
    generate_latest,
)
from starlette.responses import PlainTextResponse
from typing_extensions import Annotated

from horizon.backend.dependencies.stub import Stub
from horizon.backend.services.uow import UnitOfWork
from horizon.backend.settings import Settings
from horizon.backend.settings.server import MonitoringSettings
from horizon.backend.utils.slug import slugify

router = APIRouter(tags=["Monitoring"], prefix="/monitoring")


@router.get(
    "/stats",
    summary="Get usage stats",
    description="Return the number of users, namespaces and HWMs in database as a Prometheus metrics format",
)
async def stats(
    request: Request,
    unit_of_work: Annotated[UnitOfWork, Depends()],
    settings: Annotated[Settings, Depends(Stub(Settings))],
) -> PlainTextResponse:
    # asyncio.gather cannot be used here https://docs.sqlalchemy.org/en/20/errors.html#error-isce
    user_count_value = await unit_of_work.user.count()
    namespace_count_value = await unit_of_work.namespace.count()
    hwm_count_value = await unit_of_work.hwm.count()

    metrics = [
        (
            "horizon_user_count",
            "Number of users registered in Horizon database",
            user_count_value,
        ),
        (
            "horizon_namespace_count",
            "Number of namespaces created in Horizon database",
            namespace_count_value,
        ),
        (
            "horizon_hwm_count",
            "Number of HWMs created in Horizon database",
            hwm_count_value,
        ),
    ]

    # implementation is the same as used in starlette-exporter:
    # https://github.com/stephenhillier/starlette_exporter/blob/7b011cd/starlette_exporter/__init__.py#L44
    # https://github.com/stephenhillier/starlette_exporter/blob/7b011cd/starlette_exporter/middleware.py#L277
    app_name = slugify(request.app.title)
    registry = CollectorRegistry()
    label_keys = list(settings.server.monitoring.labels.keys())
    label_values = list(settings.server.monitoring.labels.values())

    for name, description, value in metrics:
        Gauge(
            name,
            description,
            registry=registry,
            labelnames=["app_name", *label_keys],
        ).labels(
            app_name,
            *label_values,
        ).set(value)

    headers = {"Content-Type": CONTENT_TYPE_LATEST}
    return PlainTextResponse(generate_latest(registry), headers=headers)


def apply_monitoring_stats_middleware(app: FastAPI, settings: MonitoringSettings) -> FastAPI:
    """Add monitoring stats endpoint to the application."""
    if not settings.enabled:
        return app

    app.include_router(router)
    return app



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from horizon.backend.services.current_user import current_user
from horizon.backend.services.uow import UnitOfWork



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from sqlalchemy import BigInteger, ForeignKey, String, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship

from horizon.backend.db.mixins.timestamp import TimestampMixin
from horizon.backend.db.models.base import Base
from horizon.backend.db.models.user import User


class CredentialsCache(Base, TimestampMixin):
    __tablename__ = "credentials_cache"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    # Same user can have multiple credentials (e.g. name+password, email+password)
    # So this field is named `login` instead of `username`
    login: Mapped[str] = mapped_column(
        String(256),  # noqa: WPS432
        nullable=False,
        unique=True,
        index=True,
    )
    user_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("user.id", ondelete="CASCADE"),
        nullable=False,
    )
    password_hash: Mapped[str] = mapped_column(
        Text(),
        nullable=False,
    )

    user = relationship(User, lazy="selectin")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import http
from typing import Any, NamedTuple, Type

from horizon.commons.errors.base import APIErrorSchema, BaseErrorSchema


class APIErrorResponse(NamedTuple):
    status: int
    description: str
    schema: Type[BaseErrorSchema]


_responses_by_exception: dict[type[Exception], APIErrorResponse] = {}
_responses_by_status_code: dict[int, APIErrorResponse] = {}


def register_error_response(exception: type[Exception], status: http.HTTPStatus):
    """Register mapping between exception, status code and JSON body schema."""

    def wrapper(cls):
        response = APIErrorResponse(status.value, status.phrase, cls)
        _responses_by_exception[exception] = response
        _responses_by_status_code[status.value] = response
        return cls

    return wrapper


def get_response_for_exception(exception_type: type[Exception]) -> APIErrorResponse | None:
    """Get mapping between exception type and JSON body schema (for serialization)."""
    for exception in exception_type.mro():
        if exception in _responses_by_exception:
            return _responses_by_exception[exception]
    return None


def get_response_for_status_code(status_code: int) -> APIErrorResponse | None:
    """Get mapping between status code and JSON body schema (for deserialization)."""
    return _responses_by_status_code.get(status_code, None)


def get_error_responses(
    include: set[type[BaseErrorSchema]] | None = None,
    exclude: set[type[BaseErrorSchema]] | None = None,
) -> dict[int | str, dict[str, Any]]:
    """
    Construct FastAPI responses descriptions using specified filters.
    """
    result: dict[int | str, dict[str, Any]] = {}
    for response in _responses_by_exception.values():
        if include and response.schema not in include:
            continue
        if exclude and response.schema in exclude:
            continue

        # https://fastapi.tiangolo.com/advanced/additional-responses/
        result[response.status] = {
            "model": APIErrorSchema[response.schema],  # type: ignore[name-defined]
            "description": response.description,
        }
    return result



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from fastapi import APIRouter, FastAPI
from starlette.responses import PlainTextResponse
from starlette_exporter import PrometheusMiddleware, handle_metrics

from horizon.backend.settings.server import MonitoringSettings
from horizon.backend.utils.slug import slugify

DEFAULT_SKIP_PATHS = {
    "/monitoring/metrics",
    "/monitoring/ping",
    "/monitoring/stats",
    "/static",
    "/docs",
    "/redoc",
    "/openapi.json",
    "/favicon.ico",
}

router = APIRouter(tags=["Monitoring"], prefix="/monitoring")


router.get(
    "/metrics",
    summary="Get server metrics",
    description="Return server metrics in Prometheus format, using Starlette Exporter",
    response_class=PlainTextResponse,
)(handle_metrics)


def apply_monitoring_metrics_middleware(app: FastAPI, settings: MonitoringSettings) -> FastAPI:
    """Add monitoring metrics middleware & endpoint to the application."""
    if not settings.enabled:
        return app

    skip_paths = DEFAULT_SKIP_PATHS | settings.skip_paths
    app.add_middleware(
        PrometheusMiddleware,
        app_name=slugify(app.title),
        skip_paths=skip_paths,
        **settings.dict(exclude={"enabled", "skip_paths"}),
    )
    app.include_router(router)
    return app



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from fastapi import Depends
from fastapi.security import OAuth2PasswordBearer
from typing_extensions import Annotated

from horizon.backend.db.models import User
from horizon.backend.dependencies import Stub
from horizon.backend.providers.auth import AuthProvider

oauth_schema = OAuth2PasswordBearer(tokenUrl="v1/auth/token")


async def current_user(
    auth_provider: Annotated[AuthProvider, Depends(Stub(AuthProvider))],
    auth_schema: Annotated[str, Depends(oauth_schema)],
) -> User:
    return await auth_provider.get_current_user(auth_schema)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="pydantic-orm"

from fastapi import APIRouter, Depends, status
from typing_extensions import Annotated

from horizon.backend.db.models import NamespaceUserRoleInt, User
from horizon.backend.services import UnitOfWork, current_user
from horizon.commons.errors import get_error_responses
from horizon.commons.schemas.v1 import (
    HWMBulkCopyRequestV1,
    HWMBulkDeleteRequestV1,
    HWMCreateRequestV1,
    HWMListResponseV1,
    HWMPaginateQueryV1,
    HWMResponseV1,
    HWMUpdateRequestV1,
    PageResponseV1,
)

router = APIRouter(prefix="/hwm", tags=["HWM"], responses=get_error_responses())


@router.get(
    "/",
    summary="Paginate HWM",
    dependencies=[Depends(current_user)],
)
async def paginate_hwm(
    pagination_args: Annotated[HWMPaginateQueryV1, Depends()],
    unit_of_work: Annotated[UnitOfWork, Depends()],
) -> PageResponseV1[HWMResponseV1]:
    pagination = await unit_of_work.hwm.paginate(**pagination_args.dict())
    return PageResponseV1[HWMResponseV1].from_pagination(pagination)


@router.get(
    "/{hwm_id}",
    summary="Get HWM",
    dependencies=[Depends(current_user)],
)
async def get_hwm(
    hwm_id: int,
    unit_of_work: Annotated[UnitOfWork, Depends()],
) -> HWMResponseV1:
    hwm = await unit_of_work.hwm.get(hwm_id)
    return HWMResponseV1.from_orm(hwm)


@router.post(
    "/",
    summary="Create HWM",
    status_code=status.HTTP_201_CREATED,
)
async def create_hwm(
    data: HWMCreateRequestV1,
    user: Annotated[User, Depends(current_user)],
    unit_of_work: Annotated[UnitOfWork, Depends()],
) -> HWMResponseV1:
    async with unit_of_work:
        await unit_of_work.namespace.check_user_permission(
            user_id=user.id,
            required_role=NamespaceUserRoleInt.DEVELOPER,
            namespace_id=data.namespace_id,
        )
        hwm = await unit_of_work.hwm.create(
            data=data.dict(exclude_unset=True),
            user=user,
        )
        await unit_of_work.hwm_history.create(
            hwm_id=hwm.id,
            data=hwm.to_dict(exclude={"id"}),
        )
    return HWMResponseV1.from_orm(hwm)


@router.patch(
    "/{hwm_id}",
    summary="Update HWM",
)
async def update_hwm(
    hwm_id: int,
    changes: HWMUpdateRequestV1,
    user: Annotated[User, Depends(current_user)],
    unit_of_work: Annotated[UnitOfWork, Depends()],
) -> HWMResponseV1:
    async with unit_of_work:
        hwm = await unit_of_work.hwm.get(hwm_id)
        await unit_of_work.namespace.check_user_permission(
            user_id=user.id,
            required_role=NamespaceUserRoleInt.DEVELOPER,
            namespace_id=hwm.namespace_id,
        )
        hwm = await unit_of_work.hwm.update(
            hwm_id=hwm_id,
            changes=changes.dict(exclude_unset=True),
            user=user,
        )
        await unit_of_work.hwm_history.create(
            hwm_id=hwm.id,
            data={
                **hwm.to_dict(exclude={"id"}),
                "action": "Updated",
            },
        )
    return HWMResponseV1.from_orm(hwm)


@router.delete(
    "/{hwm_id}",
    summary="Delete HWM",
    status_code=status.HTTP_204_NO_CONTENT,
)
async def delete_hwm(
    hwm_id: int,
    user: Annotated[User, Depends(current_user)],
    unit_of_work: Annotated[UnitOfWork, Depends()],
) -> None:
    async with unit_of_work:
        hwm = await unit_of_work.hwm.get(hwm_id)
        await unit_of_work.namespace.check_user_permission(
            user_id=user.id,
            required_role=NamespaceUserRoleInt.MAINTAINER,
            namespace_id=hwm.namespace_id,
        )
        hwm = await unit_of_work.hwm.delete(hwm_id=hwm_id)
        await unit_of_work.hwm_history.create(
            hwm_id=hwm.id,
            data={**hwm.to_dict(exclude={"id"}), "changed_by_user_id": user.id, "action": "Deleted"},
        )


@router.delete(
    "/",
    summary="Bulk Delete HWM",
    status_code=status.HTTP_204_NO_CONTENT,
)
async def bulk_delete_hwm(
    changes: HWMBulkDeleteRequestV1,
    user: Annotated[User, Depends(current_user)],
    unit_of_work: Annotated[UnitOfWork, Depends()],
) -> None:
    async with unit_of_work:
        await unit_of_work.namespace.check_user_permission(
            user_id=user.id,
            required_role=NamespaceUserRoleInt.MAINTAINER,
            namespace_id=changes.namespace_id,
        )
        deleted_hwms = await unit_of_work.hwm.bulk_delete(namespace_id=changes.namespace_id, hwm_ids=changes.hwm_ids)

        hwm_history_data = [
            {"hwm_id": hwm.id, "changed_by_user_id": user.id, "action": "Deleted", **hwm.to_dict(exclude={"id"})}
            for hwm in deleted_hwms
        ]

        await unit_of_work.hwm_history.bulk_create(hwm_data=hwm_history_data)


@router.post(
    "/copy",
    summary="Copy HWMs to another namespace",
    response_model=HWMListResponseV1,
    status_code=status.HTTP_201_CREATED,
)
async def copy_hwms(
    copy_request: HWMBulkCopyRequestV1,
    user: Annotated[User, Depends(current_user)],
    unit_of_work: Annotated[UnitOfWork, Depends()],
) -> HWMListResponseV1:
    async with unit_of_work:
        await unit_of_work.namespace.check_user_permission(
            user_id=user.id,
            namespace_id=copy_request.target_namespace_id,
            required_role=NamespaceUserRoleInt.DEVELOPER,
        )

        source_namespace = await unit_of_work.namespace.get(copy_request.source_namespace_id)
        copied_hwms = await unit_of_work.hwm.bulk_copy(
            source_namespace_id=source_namespace.id,
            target_namespace_id=copy_request.target_namespace_id,
            hwm_ids=copy_request.hwm_ids,
            with_history=copy_request.with_history,
        )

        hwm_history_data = []
        for copied_hwm in copied_hwms:
            history_record = {
                **copied_hwm.to_dict(exclude={"id", "changed_by_user_id"}),
                "hwm_id": copied_hwm.id,
                "action": f"Copied from namespace {copy_request.source_namespace_id} "
                f"to namespace {copy_request.target_namespace_id}",
                "changed_by_user_id": user.id,
            }
            hwm_history_data.append(history_record)

        await unit_of_work.hwm_history.bulk_create(hwm_history_data)
        return HWMListResponseV1(hwms=[HWMResponseV1.from_orm(hwm) for hwm in copied_hwms])



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from horizon.commons.exceptions.auth import AuthorizationError
from horizon.commons.exceptions.bad_request import BadRequestError
from horizon.commons.exceptions.base import ApplicationError
from horizon.commons.exceptions.entity import (
    EntityAlreadyExistsError,
    EntityNotFoundError,
)
from horizon.commons.exceptions.permission import PermissionDeniedError
from horizon.commons.exceptions.service import ServiceError

__all__ = [
    "AuthorizationError",
    "ApplicationError",
    "EntityAlreadyExistsError",
    "PermissionDeniedError",
    "BadRequestError",
    "EntityNotFoundError",
    "ServiceError",
]



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from functools import partial

from fastapi import FastAPI
from fastapi.openapi.docs import (
    get_redoc_html,
    get_swagger_ui_html,
    get_swagger_ui_oauth2_redirect_html,
)
from fastapi.openapi.utils import get_openapi
from starlette.requests import Request
from starlette.responses import JSONResponse

from horizon.backend.settings.server.openapi import OpenAPISettings


async def custom_openapi(request: Request) -> JSONResponse:
    app: FastAPI = request.app
    root_path = request.scope.get("root_path", "").rstrip("/")
    server_urls = set(filter(None, (server_data.get("url") for server_data in app.servers)))

    if root_path not in server_urls:
        if root_path and app.root_path_in_servers:
            app.servers.insert(0, {"url": root_path})
            server_urls.add(root_path)

    return JSONResponse(app.openapi())


def custom_openapi_schema(app: FastAPI, settings: OpenAPISettings) -> dict:
    if app.openapi_schema:
        return app.openapi_schema

    openapi_schema = get_openapi(
        title=app.title,
        version=app.version,
        openapi_version=app.openapi_version,
        summary=app.summary,
        description=app.description,
        terms_of_service=app.terms_of_service,
        contact=app.contact,
        license_info=app.license_info,
        routes=app.routes,
        webhooks=app.webhooks.routes,
        tags=app.openapi_tags,
        servers=app.servers,
        separate_input_output_schemas=app.separate_input_output_schemas,
    )
    # https://redocly.com/docs/api-reference-docs/specification-extensions/x-logo/
    openapi_schema["info"]["x-logo"] = {
        "url": str(settings.logo.url),
        "altText": str(settings.logo.alt_text),
        "backgroundColor": f"#{settings.logo.background_color}",  # noqa: WPS237
        "href": str(settings.logo.href),
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


async def custom_swagger_ui_html(request: Request):
    app: FastAPI = request.app
    settings: OpenAPISettings = app.state.settings.server.openapi
    root_path = request.scope.get("root_path", "").rstrip("/")
    openapi_url = root_path + request.app.openapi_url  # type: ignore[arg-type]
    return get_swagger_ui_html(
        openapi_url=openapi_url,
        title=f"{app.title} - Swagger UI",
        oauth2_redirect_url=app.swagger_ui_oauth2_redirect_url,
        swagger_js_url=str(settings.swagger.js_url),
        swagger_css_url=str(settings.swagger.css_url),
        swagger_favicon_url=str(settings.favicon.url),
    )


async def custom_swagger_ui_redirect(request: Request):
    return get_swagger_ui_oauth2_redirect_html()


async def custom_redoc_html(request: Request):
    app: FastAPI = request.app
    settings: OpenAPISettings = app.state.settings.server.openapi
    root_path = request.scope.get("root_path", "").rstrip("/")
    openapi_url = root_path + request.app.openapi_url  # type: ignore[arg-type]
    return get_redoc_html(
        openapi_url=openapi_url,
        title=f"{app.title} - ReDoc",
        redoc_js_url=settings.redoc.js_url,
        redoc_favicon_url=settings.favicon.url,
        with_google_fonts=False,
    )


def apply_openapi_middleware(app: FastAPI, settings: OpenAPISettings) -> FastAPI:
    """Add OpenAPI middleware to the application."""
    # https://fastapi.tiangolo.com/how-to/custom-docs-ui-assets/#include-the-custom-docs
    if settings.enabled:
        app.openapi_url = "/openapi.json"
        app.add_route(app.openapi_url, custom_openapi, include_in_schema=False)

    if settings.swagger.enabled:
        app.docs_url = "/docs"
        app.swagger_ui_oauth2_redirect_url = "/docs/oauth2-redirect"
        app.add_route(app.docs_url, custom_swagger_ui_html, include_in_schema=False)
        app.add_route(app.swagger_ui_oauth2_redirect_url, custom_swagger_ui_redirect, include_in_schema=False)

    if settings.redoc.enabled:
        app.redoc_url = "/redoc"
        app.add_route(app.redoc_url, custom_redoc_html, include_in_schema=False)

    app.openapi = partial(custom_openapi_schema, app=app, settings=settings)  # type: ignore[method-assign]
    return app



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, TypeVar

from pydantic import AnyHttpUrl

Session = TypeVar("Session", bound=Any)


class BaseAuth(ABC):
    """
    Base authentication class.
    """

    type: str

    @abstractmethod
    def patch_session(self, session: Session) -> Session:
        """Get session object and return it with necessary patches applied"""
        ...

    @abstractmethod
    def fetch_token_kwargs(self, base_url: AnyHttpUrl) -> dict[str, Any]:
        """Return key-values arguments for ``client.fetch_token(...)`` method.

        Empty dict means that this method should not be called.
        """
        ...



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from fastapi import Depends
from sqlalchemy.ext.asyncio import AsyncSession
from typing_extensions import Annotated

from horizon.backend.db.repositories import (
    CredentialsCacheRepository,
    HWMHistoryRepository,
    HWMRepository,
    NamespaceHistoryRepository,
    NamespaceRepository,
    UserRepository,
)
from horizon.backend.dependencies import Stub


class UnitOfWork:
    def __init__(
        self,
        session: Annotated[AsyncSession, Depends(Stub(AsyncSession))],
    ):
        self._session = session
        self.namespace = NamespaceRepository(session=session)
        self.hwm_history = HWMHistoryRepository(session=session)
        self.namespace_history = NamespaceHistoryRepository(session=session)
        self.user = UserRepository(session=session)
        self.hwm = HWMRepository(session=session)
        self.credentials_cache = CredentialsCacheRepository(session=session)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            await self._session.rollback()
        else:
            await self._session.commit()



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
import http
from typing import Any

from pydantic import BaseModel
from typing_extensions import Literal

from horizon.commons.errors.base import BaseErrorSchema
from horizon.commons.errors.registration import register_error_response
from horizon.commons.exceptions.entity import EntityAlreadyExistsError


class AlreadyExistsDetailsSchema(BaseModel):
    entity_type: str
    field: str
    value: Any


@register_error_response(
    exception=EntityAlreadyExistsError,
    status=http.HTTPStatus.CONFLICT,
)
class AlreadyExistsSchema(BaseErrorSchema):
    code: Literal["already_exists"] = "already_exists"
    details: AlreadyExistsDetailsSchema

    def to_exception(self) -> EntityAlreadyExistsError:
        return EntityAlreadyExistsError(
            entity_type=self.details.entity_type,
            field=self.details.field,
            value=self.details.value,
        )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field
from pydantic import __version__ as pydantic_version

from horizon.commons.schemas.v1.pagination import PaginateQueryV1


class NamespaceHistoryPaginateQueryV1(PaginateQueryV1):
    """Query params for Namespace pagination request."""

    namespace_id: int = Field(description="Namespace id")

    # more arguments can be added in future


class NamespaceHistoryResponseV1(BaseModel):
    """Namespace history response."""

    id: int = Field(description="Namespace history item id")
    namespace_id: int = Field(description="Namespace id history is bound to")
    name: str = Field(description="Namespace name")
    description: str = Field(description="Namespace description")
    owned_by: Optional[str] = Field(default=None, description="The namespace owner")
    action: str = Field(description="Action performed on the namespace record")
    changed_at: datetime = Field(description="Timestamp of a change of the namespace data")
    changed_by: Optional[str] = Field(default=None, description="User who changed the namespace data")

    class Config:
        if pydantic_version >= "2":
            from_attributes = True
        else:
            orm_mode = True



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import http
import logging

from asgi_correlation_id import correlation_id
from fastapi import HTTPException, Request, Response
from fastapi.exceptions import RequestValidationError
from pydantic import ValidationError

from horizon.backend.settings.server import ServerSettings
from horizon.commons.errors.base import APIErrorSchema, BaseErrorSchema
from horizon.commons.errors.registration import get_response_for_exception
from horizon.commons.exceptions import ApplicationError, ServiceError

logger = logging.getLogger(__name__)


def http_exception_handler(_request: Request, exc: HTTPException) -> Response:
    content = BaseErrorSchema(
        code=http.HTTPStatus(exc.status_code).name.lower(),
        message=exc.detail,
        details=None,
    )
    return exception_json_response(
        status=exc.status_code,
        content=content,
        headers=exc.headers,
    )


def unknown_exception_handler(request: Request, exc: Exception) -> Response:
    logger.exception("Got unhandled error: %s", exc, exc_info=exc)

    server: ServerSettings = request.app.state.settings.server
    details = None
    if request.app.debug:
        details = exc.args

    content = BaseErrorSchema(
        code="unknown",
        message="Got unhandled exception. Please contact support",
        details=details,
    )
    return exception_json_response(
        status=http.HTTPStatus.INTERNAL_SERVER_ERROR.value,
        content=content,
        # https://github.com/snok/asgi-correlation-id#exception-handling
        headers={server.request_id.header_name: correlation_id.get() or ""},
    )


def service_exception_handler(request: Request, exc: ServiceError) -> Response:
    logger.exception("Got service error: %s", exc, exc_info=exc)

    server: ServerSettings = request.app.state.settings.server
    details = None
    if request.app.debug:
        details = exc.message

    content = BaseErrorSchema(
        code="service_unavailable",
        message="Service unavailable",
        details=details,
    )
    return exception_json_response(
        status=http.HTTPStatus.SERVICE_UNAVAILABLE.value,
        content=content,
        # https://github.com/snok/asgi-correlation-id#exception-handling
        headers={server.request_id.header_name: correlation_id.get() or ""},
    )


def validation_exception_handler(request: Request, exc: RequestValidationError) -> Response:
    response = get_response_for_exception(ValidationError)
    if not response:
        return unknown_exception_handler(request, exc)

    # code and message is set within class implementation
    errors = []
    for error in exc.errors():
        # pydantic Error classes are not serializable, drop it
        error.get("ctx", {}).pop("error", None)
        errors.append(error)

    content = response.schema(  # type: ignore[call-arg]
        details=errors,
    )
    return exception_json_response(
        status=response.status,
        content=content,
    )


def application_exception_handler(request: Request, exc: ApplicationError) -> Response:
    response = get_response_for_exception(type(exc))
    if not response:
        return unknown_exception_handler(request, exc)

    logger.error("%s", exc, exc_info=logger.isEnabledFor(logging.DEBUG))

    # code is set within class implementation
    content = response.schema(  # type: ignore[call-arg]
        message=exc.message,
        details=exc.details,
    )
    return exception_json_response(
        status=response.status,
        content=content,
    )


def exception_json_response(
    status: int,
    content: BaseErrorSchema,
    headers: dict[str, str] | None = None,
) -> Response:
    # Using Response + `model.json()` because JSONResponse + `model.dict()` does not convert Unset() to JSON value "<unset>"
    content_type = type(content)
    error_schema = APIErrorSchema[content_type]  # type: ignore[valid-type]
    return Response(
        status_code=status,
        content=error_schema(error=content).json(by_alias=True),
        media_type="application/json",
        headers=headers,
    )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

"""
Settings for LDAPAuthProvider class.

Basic LDAP terminology is explained here: `LDAP Overview <https://www.zytrax.com/books/ldap/ch2/>`_
"""

import textwrap
from typing import TYPE_CHECKING, Optional, Type, Union

from bonsai import LDAPSearchScope
from pydantic import AnyUrl, BaseModel, Field, SecretStr
from pydantic import __version__ as pydantic_version
from pydantic import validator
from typing_extensions import Annotated, Literal

from horizon.backend.settings.auth.jwt import JWTSettings

if TYPE_CHECKING:
    LDAPUrl = AnyUrl
elif pydantic_version < "2":

    class LDAPUrl(AnyUrl):  # noqa: WPS440
        """LDAP connection url, like ``ldap://127.0.0.1:389`` or ``ldaps://127.0.0.1:636``"""

        allowed_schemes = ["ldap", "ldaps"]
        host_required = True

else:
    from pydantic import UrlConstraints

    LDAPUrl: Type[AnyUrl] = Annotated[  # noqa: WPS440
        AnyUrl,
        UrlConstraints(allowed_schemes=["ldap", "ldaps"], host_required=True),
    ]


class LDAPCredentials(BaseModel):
    """LDAP lookup query is executed using this credentials
    (instead of login and password provided by user).

    Examples
    --------

    .. code-block:: bash

        HORIZON__AUTH__LDAP__LOOKUP__CREDENTIALS__USER=uid=techuser,ou=users,dc=example,dc=com
        HORIZON__AUTH__LDAP__LOOKUP__CREDENTIALS__PASSWORD=somepassword
    """

    user: str = Field(
        description="DN of user which is used for calling ``lookup`` query in LDAP",
    )
    password: SecretStr = Field(
        description="This user password",
    )


class LDAPConnectionPoolSettings(BaseModel):
    """Settings related to LDAP connection pool.

    Examples
    --------

    .. code-block:: bash

        HORIZON__AUTH__LDAP__LOOKUP__POOL__ENABLED=True
        HORIZON__AUTH__LDAP__LOOKUP__POOL__MAX=10
        HORIZON__AUTH__LDAP__LOOKUP__POOL__CHECK_ON_STARTUP=True
    """

    enabled: bool = Field(
        default=True,
        description="Set to ``True`` to enable connection pool",
    )
    initial: int = Field(
        default=1,
        description="Initial size of connection pool",
    )
    max: int = Field(
        default=10,
        description="Maximum size of connection pool",
    )


class LDAPLookupSettings(BaseModel):
    """Settings related to LDAP lookup.

    Examples
    --------

    .. code-block:: bash

        HORIZON__AUTH__LDAP__LOOKUP__ENABLED=True
        HORIZON__AUTH__LDAP__LOOKUP__POOL__ENABLED=True
        HORIZON__AUTH__LDAP__LOOKUP__CREDENTIALS__USER=uid=techuser,ou=users,dc=example,dc=com
        HORIZON__AUTH__LDAP__LOOKUP__CREDENTIALS__PASSWORD=somepassword
        HORIZON__AUTH__LDAP__LOOKUP__QUERY=(uid={login})
    """

    enabled: bool = Field(
        default=True,
        description="Set to ``True`` to enable lookup",
    )
    check_on_startup: bool = Field(
        default=True,
        description="If ``True``, and LDAP is not available during application start, abort application startup",
    )
    pool: LDAPConnectionPoolSettings = Field(
        default_factory=LDAPConnectionPoolSettings,
        description="LDAP connection pool settings",
    )
    credentials: Optional[LDAPCredentials] = Field(
        default=None,
        description="Credentials used for connecting to LDAP while performing user lookup",
    )
    query_template: str = Field(
        default="({uid_attribute}={login})",
        description=textwrap.dedent(
            """
            LDAP query send in lookup request.

            Usually lookup is performed against attributes ``uid`` (LDAP) or ``sAMAccountName`` (ActiveDirectory).
            You can also pass any query string supported by LDAP.
            See `Bonsai documentation <https://bonsai.readthedocs.io/en/latest/tutorial.html#searching>`_.

            Supported substitution values (see :obj:`horizon.backend.settings.auth.ldap.LDAPSettings`.):
              * ``{uid_attribute}``
              * ``{login}``
            """,
        ),
    )
    scope: LDAPSearchScope = Field(
        default=LDAPSearchScope.ONELEVEL,
        description=textwrap.dedent(
            """
            Lookup scope. Use ``SUBTREE`` for ActiveDirectory.

            See `Bonsai documentation <https://bonsai.readthedocs.io/en/latest/api.html#bonsai.LDAPSearchScope.ONE>`_.
            """,
        ),
    )

    @validator("scope", pre=True)
    def _convert_scope_to_enum(cls, value: Union[str, int, LDAPSearchScope]) -> LDAPSearchScope:
        if isinstance(value, str):
            return LDAPSearchScope[value.upper()]
        return LDAPSearchScope(value)


class LDAPSettings(BaseModel):
    """Settings related to LDAP interaction.

    Examples
    --------

    .. code-block:: bash

        HORIZON__AUTH__LDAP__URL=ldap://ldap.domain.com:389
        HORIZON__AUTH__LDAP__UID_ATTRIBUTE=sAMAccountName
    """

    url: LDAPUrl = Field(
        description="LDAP URL to connect to",
    )
    timeout_seconds: Optional[int] = Field(
        default=10,
        description="LDAP request timeout, in seconds. ``None`` means no timeout",
    )
    auth_mechanism: Literal["SIMPLE", "DIGEST-MD5"] = Field(
        default="SIMPLE",
        description="LDAP auth mechanism, used for ``bind`` request",
    )
    base_dn: str = Field(
        description="Organization DN, e.g. ``ou=users,dc=example,dc=com``",
    )
    uid_attribute: str = Field(
        default="uid",
        description=textwrap.dedent(
            """
            Attribute containing username.

            Usually ``uid`` (LDAP) or ``sAMAccountName`` (ActiveDirectory).
            """,
        ),
    )
    bind_dn_template: str = Field(
        default="{uid_attribute}={login},{base_dn}",
        description=textwrap.dedent(
            """
            Template for building DN string, used for checking credentials in LDAP. You can pass any DN value supported by LDAP.

            Supported substitution values:
              * ``{login}``
              * ``{uid_attribute}`` (see :obj:`~uid_attribute`)
              * ``{base_dn}`` (see :obj:`~base_dn`)
            """,
        ),
    )

    lookup: LDAPLookupSettings = Field(
        default_factory=LDAPLookupSettings,
        description="LDAP search options",
    )


class LDAPAuthProviderSettings(BaseModel):
    """Settings for LDAPAuthProvider.

    Examples
    --------

    .. code-block:: bash

        HORIZON__AUTH__PROVIDER=horizon.backend.providers.auth.ldap.LDAPAuthProvider
        HORIZON__AUTH__ACCESS_KEY__SECRET_KEY=secret
        HORIZON__AUTH__LDAP__URL=ldap://ldap.domain.com:389
        HORIZON__AUTH__LDAP__LOOKUP__ENABLED=True
        HORIZON__AUTH__LDAP__LOOKUP__POOL__ENABLED=True
        HORIZON__AUTH__LDAP__LOOKUP__CREDENTIALS__USER=uid=techuser,ou=users,dc=example,dc=com
        HORIZON__AUTH__LDAP__LOOKUP__CREDENTIALS__PASSWORD=somepassword
    """

    access_token: JWTSettings = Field(description="Access-token related settings")
    ldap: LDAPSettings = Field(description="LDAP related settings")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from typing import Type

from fastapi import FastAPI, HTTPException
from fastapi.exceptions import RequestValidationError
from sqlalchemy.ext.asyncio import AsyncSession, async_engine_from_config

import horizon
from horizon.backend.api.handlers import (
    application_exception_handler,
    http_exception_handler,
    service_exception_handler,
    unknown_exception_handler,
    validation_exception_handler,
)
from horizon.backend.api.router import api_router
from horizon.backend.db.factory import create_session_factory
from horizon.backend.middlewares import apply_middlewares
from horizon.backend.providers.auth.base import AuthProvider
from horizon.backend.settings import Settings
from horizon.commons.exceptions import ApplicationError, ServiceError


def application_factory(settings: Settings) -> FastAPI:
    application = FastAPI(
        title="Horizon",
        description="Horizon is an application that implements simple HWM Store",
        version=horizon.__version__,
        debug=settings.server.debug,
        # will be set up by middlewares
        openapi_url=None,
        docs_url=None,
        redoc_url=None,
    )

    application.state.settings = settings
    application.include_router(api_router)

    application.add_exception_handler(ServiceError, service_exception_handler)  # type: ignore[arg-type]
    application.add_exception_handler(ApplicationError, application_exception_handler)  # type: ignore[arg-type]
    application.add_exception_handler(
        RequestValidationError,
        validation_exception_handler,  # type: ignore[arg-type]
    )
    application.add_exception_handler(HTTPException, http_exception_handler)  # type: ignore[arg-type]
    application.add_exception_handler(Exception, unknown_exception_handler)

    engine = async_engine_from_config(settings.database.dict(), prefix="")
    session_factory = create_session_factory(engine)

    application.dependency_overrides.update(
        {
            Settings: lambda: settings,
            AsyncSession: session_factory,  # type: ignore[dict-item]
        },
    )

    # get AuthProvider class from settings, and perform setup
    auth_class: Type[AuthProvider] = settings.auth.provider  # type: ignore[assignment]
    auth_class.setup(application)

    apply_middlewares(application, settings)
    return application


def get_application():
    settings = Settings()
    return application_factory(settings=settings)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from horizon.backend.middlewares.monitoring.metrics import (
    apply_monitoring_metrics_middleware,
)
from horizon.backend.middlewares.monitoring.stats import (
    apply_monitoring_stats_middleware,
)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass
class Pagination(Generic[T]):
    items: list[T]
    page: int
    page_size: int
    total_count: int

    @property
    def has_previous(self) -> bool:
        return self.page > 1

    @property
    def previous_page(self) -> int | None:
        if self.has_previous:
            return self.page - 1
        return None

    @property
    def has_next(self) -> bool:
        previous_items = (self.page - 1) * self.page_size
        return previous_items + len(self.items) < self.total_count

    @property
    def next_page(self) -> int | None:
        if self.has_next:
            return self.page + 1
        return None

    @property
    def pages_count(self) -> int:
        return int(math.ceil(self.total_count / float(self.page_size))) or 1



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
import http
from typing import Any, List, Union

from pydantic import BaseModel, Field, ValidationError
from pydantic import __version__ as pydantic_version
from typing_extensions import Literal

from horizon.commons.errors.base import BaseErrorSchema
from horizon.commons.errors.registration import register_error_response


class InvalidRequestBaseErrorSchema(BaseModel):
    loc: List[Union[str, int]] = Field(alias="location")
    msg: str = Field(alias="message")
    type: str = Field(alias="code")

    if pydantic_version >= "2":
        ctx: dict = Field(default_factory=dict, alias="context")
        input: Any = Field(default=None)

    class Config:
        if pydantic_version >= "2":
            populate_by_name = True
        else:
            allow_population_by_field_name = True


@register_error_response(
    exception=ValidationError,
    status=http.HTTPStatus.UNPROCESSABLE_ENTITY,
)
class InvalidRequestSchema(BaseErrorSchema):
    code: Literal["invalid_request"] = "invalid_request"
    message: Literal["Invalid request"] = "Invalid request"
    details: List[InvalidRequestBaseErrorSchema]



#!/bin/env python3
# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

import json
import sys

from fastapi import FastAPI

from horizon.backend import get_application


def get_openapi_schema(app: FastAPI) -> dict:
    return app.openapi()


if __name__ == "__main__":
    app = get_application()
    schema = get_openapi_schema(app)
    file_path = sys.argv[1]
    with open(file_path, "w") as file:
        json.dump(schema, file)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

import logging
from logging.config import dictConfig
from pathlib import Path

import yaml  # type: ignore

logger = logging.getLogger(__name__)


class LoggingSetupError(Exception):
    pass


def setup_logging(config_path: Path) -> None:
    """Parse file with logging configuration, and setup logging accordingly"""
    if not config_path.exists():
        raise OSError(f"Logging configuration file '{config_path}' does not exist")

    try:
        config = yaml.safe_load(config_path.read_text())
        dictConfig(config)
    except Exception as e:
        raise LoggingSetupError(f"Error reading logging configuration '{config_path}'") from e



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from fastapi import FastAPI

from horizon.backend.middlewares.application_version import (
    apply_application_version_middleware,
)
from horizon.backend.middlewares.cors import apply_cors_middleware
from horizon.backend.middlewares.logging import setup_logging
from horizon.backend.middlewares.monitoring import (
    apply_monitoring_metrics_middleware,
    apply_monitoring_stats_middleware,
)
from horizon.backend.middlewares.openapi import apply_openapi_middleware
from horizon.backend.middlewares.request_id import apply_request_id_middleware
from horizon.backend.middlewares.static_files import apply_static_files
from horizon.backend.settings import Settings


def apply_middlewares(
    application: FastAPI,
    settings: Settings,
) -> FastAPI:
    """Add middlewares to the application."""

    if settings.server.logging.setup:
        setup_logging(settings.server.logging.get_log_config_path())

    apply_cors_middleware(application, settings.server.cors)
    apply_monitoring_metrics_middleware(application, settings.server.monitoring)
    apply_monitoring_stats_middleware(application, settings.server.monitoring)
    apply_request_id_middleware(application, settings.server.request_id)
    apply_application_version_middleware(application, settings.server.application_version)
    apply_openapi_middleware(application, settings.server.openapi)
    apply_static_files(application, settings.server.static_files)

    return application



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from authlib.oauth2.auth import OAuth2Token as AuthlibToken
from jose import ExpiredSignatureError, jwt
from pydantic import AnyHttpUrl, BaseModel, validator
from typing_extensions import Literal

from horizon.client.auth.base import BaseAuth, Session


class AccessToken(BaseAuth, BaseModel):
    """Authorization using access token.

    Token is passed in ``Authorization: Bearer ${token}`` header,
    and does not support refreshing.

    Parameters
    ----------
    token: str
        Access token

    Examples
    --------

    >>> from horizon.client.auth import AccessToken
    >>> auth = AccessToken(token="my.access.token")
    """

    token: str

    type: Literal["access_token"] = "access_token"

    def patch_session(self, session: Session) -> Session:
        if session.token:
            return session

        # https://github.com/lepture/authlib/issues/600
        token_decoded = jwt.decode(self.token, key="NONE", options={"verify_signature": False})
        session.token = AuthlibToken.from_dict(
            {
                "access_token": self.token,
                "token_type": "Bearer",
                "expires_at": token_decoded["exp"],
            },
        )
        return session

    def fetch_token_kwargs(self, base_url: AnyHttpUrl) -> dict[str, str]:
        return {}

    @validator("token")
    def _validate_token(cls, value):
        token_decoded = jwt.decode(value, key="NONE", options={"verify_signature": False})
        if "exp" not in token_decoded:
            raise ExpiredSignatureError("Missing expiration time in token")
        return value



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import Any

from horizon.commons.exceptions.base import ApplicationError


class PermissionDeniedError(ApplicationError):
    """
    Permission denied for performing the requested action.

    Examples
    --------

    >>> from horizon.commons.exceptions import PermissionDeniedError
    >>> raise PermissionDeniedError(required_role="DEVELOPER", actual_role="GUEST")
    Traceback (most recent call last):
    horizon.commons.exceptions.PermissionDeniedError: Permission denied. User has role GUEST but action requires at least DEVELOPER.
    """

    required_role: str
    """Required role to perform action"""

    actual_role: str
    """Actual user role"""

    def __init__(self, required_role: str, actual_role: str):
        self.required_role = required_role
        self.actual_role = actual_role

    @property
    def message(self) -> str:
        return f"Permission denied. User has role {self.actual_role} but action requires at least {self.required_role}."

    @property
    def details(self) -> dict[str, Any]:
        return {
            "required_role": self.required_role,
            "actual_role": self.actual_role,
        }



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from fastapi import APIRouter

from horizon.backend.api.v1.router.auth import router as auth_router
from horizon.backend.api.v1.router.hwm import router as hwm_router
from horizon.backend.api.v1.router.hwm_history import router as hwm_history_router
from horizon.backend.api.v1.router.namespace_history import (
    router as namespace_history_router,
)
from horizon.backend.api.v1.router.namespaces import router as namespaces_router
from horizon.backend.api.v1.router.users import router as users_router

router = APIRouter(prefix="/v1")
router.include_router(auth_router)
router.include_router(users_router)
router.include_router(namespaces_router)
router.include_router(hwm_router)
router.include_router(hwm_history_router)
router.include_router(namespace_history_router)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from typing import List

from pydantic import BaseModel, Field


class CORSSettings(BaseModel):
    """CORS Middleware Settings.

    See `CORSMiddleware <https://www.starlette.io/middleware/#corsmiddleware>`_ documentation.

    .. note::

        You can pass here any extra option supported by ``CORSMiddleware``,
        even if it is not mentioned in documentation.

    Examples
    --------

    For development environment:

    .. code-block:: bash

        HORIZON__SERVER__CORS__ENABLED=True
        HORIZON__SERVER__CORS__ALLOW_ORIGINS=["*"]
        HORIZON__SERVER__CORS__ALLOW_METHODS=["*"]
        HORIZON__SERVER__CORS__ALLOW_HEADERS=["*"]
        HORIZON__SERVER__CORS__EXPOSE_HEADERS=["X-Request-ID"]

    For production environment:

    .. code-block:: bash

        HORIZON__SERVER__CORS__ENABLED=True
        HORIZON__SERVER__CORS__ALLOW_ORIGINS=["production.example.com"]
        HORIZON__SERVER__CORS__ALLOW_METHODS=["GET"]
        HORIZON__SERVER__CORS__ALLOW_HEADERS=["X-Request-ID", "X-Request-With"]
        HORIZON__SERVER__CORS__EXPOSE_HEADERS=["X-Request-ID"]
        # custom option passed directly to middleware
        HORIZON__SERVER__CORS__MAX_AGE=600
    """

    enabled: bool = Field(default=True, description="Set to ``True`` to enable middleware")
    allow_origins: List[str] = Field(default_factory=list, description="Domains allowed for CORS")
    allow_credentials: bool = Field(
        default=False,
        description="If ``True``, cookies should be supported for cross-origin request",
    )
    allow_methods: List[str] = Field(default=["GET"], description="HTTP Methods allowed for CORS")
    # https://github.com/snok/asgi-correlation-id#cors
    allow_headers: List[str] = Field(
        default=["X-Request-ID", "X-Request-With"],
        description="HTTP headers allowed for CORS",
    )
    expose_headers: List[str] = Field(default=["X-Request-ID"], description="HTTP headers exposed from backend")

    class Config:
        extra = "allow"



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="pydantic-orm"

from typing import Union

from fastapi import APIRouter, Depends
from typing_extensions import Annotated

from horizon.backend.db.models import User
from horizon.backend.services import current_user
from horizon.commons.errors import get_error_responses
from horizon.commons.schemas.v1 import UserResponseV1, UserResponseV1WithAdmin

router = APIRouter(prefix="/users", tags=["Users"])


@router.get("/me", summary="Get current user info", responses=get_error_responses())
async def whoami(
    user: Annotated[User, Depends(current_user)],
) -> Union[UserResponseV1WithAdmin, UserResponseV1]:
    if user.is_admin:
        return UserResponseV1WithAdmin.from_orm(user)
    return UserResponseV1.from_orm(user)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from abc import ABC
from typing import Generic, TypeVar

from sqlalchemy import ScalarResult, Select, delete, func, insert, select, update
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.sql import ColumnElement, SQLColumnExpression
from sqlalchemy.sql.dml import ReturningDelete, ReturningInsert, ReturningUpdate

from horizon.backend.db.models import Base
from horizon.commons.dto.pagination import Pagination

Model = TypeVar("Model", bound=Base)


class Repository(ABC, Generic[Model]):
    def __init__(
        self,
        session: AsyncSession,
    ) -> None:
        self._session = session

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}()"

    @classmethod
    def model_type(cls) -> type[Model]:
        # Get `User` from `UserRepository(Repository[User])`
        return cls.__orig_bases__[0].__args__[0]  # type: ignore[attr-defined]

    async def _get_by_id(
        self,
        id: int,
        *where: ColumnElement,
    ) -> Model | None:
        model_type = self.model_type()
        query: Select = select(model_type).filter_by(id=id)
        if where:
            query = query.where(*where)

        return await self._session.scalar(query)

    async def _get(
        self,
        *where: ColumnElement,
    ) -> Model | None:
        model_type = self.model_type()
        query: Select = select(model_type).where(*where)
        return await self._session.scalar(query)

    async def _create(
        self,
        data: dict,
    ) -> Model:
        model_type = self.model_type()
        query: ReturningInsert[tuple[Model]] = insert(model_type).values(**data).returning(model_type)
        result = await self._session.scalars(query)
        return result.one()

    async def _update(
        self,
        where: list[ColumnElement],
        changes: dict,
    ) -> Model | None:
        model_type = self.model_type()
        query: ReturningUpdate[tuple[Model]] = update(model_type).where(*where).values(**changes).returning(model_type)
        return await self._session.scalar(query)

    async def _delete(self, id: int) -> Model | None:
        model_type = self.model_type()
        query: ReturningDelete[tuple[Model]] = (
            delete(model_type).where(model_type.id == id).returning(model_type)  # type: ignore[attr-defined]
        )
        return await self._session.scalar(query)

    async def _paginate(
        self,
        order_by: list[SQLColumnExpression],
        page: int,
        page_size: int,
        where: list[SQLColumnExpression] | None = None,
    ) -> Pagination[Model]:
        model_type = self.model_type()
        query: Select = select(model_type)
        if where:
            query = query.where(*where)

        items_result: ScalarResult[Model] = await self._session.scalars(
            query.order_by(*order_by).limit(page_size).offset((page - 1) * page_size),
        )
        total_count: int = await self._session.scalar(  # type: ignore[assignment]
            select(func.count()).select_from(query.subquery()),
        )
        return Pagination[model_type](  # type: ignore[valid-type]
            items=list(items_result.all()),
            total_count=total_count,
            page=page,
            page_size=page_size,
        )

    async def _count(
        self,
        where: list[SQLColumnExpression] | None = None,
    ) -> int:
        model_type = self.model_type()
        query: Select = select(func.count()).select_from(model_type)
        if where:
            query = query.where(*where)

        result = await self._session.scalars(query)
        return result.one()



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
"""Add hwm table

Revision ID: edd2e353ca38
Revises: e29b66594970
Create Date: 2023-10-16 19:10:26.499633

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "edd2e353ca38"
down_revision = "e29b66594970"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "hwm",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("namespace_id", sa.BigInteger(), nullable=True),
        sa.Column("name", sa.String(length=256), nullable=False),
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("type", sa.String(length=64), nullable=False),
        sa.Column("value", sa.JSON(), nullable=False),
        sa.Column("entity", sa.Text(), nullable=True),
        sa.Column("expression", sa.Text(), nullable=True),
        sa.Column("changed_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=False),
        sa.Column("changed_by_user_id", sa.BigInteger(), nullable=True),
        sa.Column("is_deleted", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["changed_by_user_id"],
            ["user.id"],
            name=op.f("fk__hwm__changed_by_user_id__user"),
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["namespace_id"],
            ["namespace.id"],
            name=op.f("fk__hwm__namespace_id__namespace"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__hwm")),
        sa.UniqueConstraint("namespace_id", "name", name="hwm_name_unique_per_namespace"),
    )
    op.create_index(op.f("ix__hwm__name"), "hwm", ["name"], unique=False)
    op.create_index(op.f("ix__hwm__namespace_id"), "hwm", ["namespace_id"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix__hwm__namespace_id"), table_name="hwm")
    op.drop_index(op.f("ix__hwm__name"), table_name="hwm")
    op.drop_table("hwm")
    # ### end Alembic commands ###



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from pydantic import BaseModel, Field
from pydantic import __version__ as pydantic_version


class UserResponseV1(BaseModel):
    """User info response."""

    id: int = Field(description="Internal user id, not for external usage")
    username: str = Field(description="User name, unique in the entire database")

    class Config:
        if pydantic_version >= "2":
            from_attributes = True
        else:
            orm_mode = True


class UserResponseV1WithAdmin(UserResponseV1):
    """Extended user info response including is_admin."""

    is_admin: bool = Field(description="Indicates if the user is a superadmin")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

import textwrap

from pydantic import BaseModel, Field, SecretStr


class JWTSettings(BaseModel):
    """Settings related to JWT tokens.

    Examples
    --------

    .. code-block:: bash

        HORIZON__AUTH__ACCESS_KEY__SECRET_KEY=somesecret
        HORIZON__AUTH__ACCESS_KEY__EXPIRE_SECONDS=3600  # 1 hour
    """

    secret_key: SecretStr = Field(
        description=textwrap.dedent(
            """
            Secret key for signing JWT tokens.

            Can be any string. It is recommended to generate random value for every application instance.
            """,
        ),
    )
    security_algorithm: str = Field(
        default="HS256",
        description=textwrap.dedent(
            """
            Algorithm used for signing JWT tokens.

            See `python-jose <https://python-jose.readthedocs.io/en/latest/jws/index.html#supported-algorithms>`_
            documentation.
            """,
        ),
    )
    expire_seconds: int = Field(
        default=10 * 60 * 60,
        description="Token expiration time, in seconds",
    )



#!/bin/env python3

# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import argparse
import asyncio
import logging

from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.future import select

from horizon.backend.db.models import User
from horizon.backend.middlewares import setup_logging
from horizon.backend.settings import Settings


async def add_admins(session: AsyncSession, usernames: list[str]) -> None:
    logging.info("Adding SUPERADMIN users:")
    result = await session.execute(select(User).where(User.username.in_(usernames)).order_by(User.username))
    users = result.scalars().all()

    not_found = set(usernames)
    for user in users:
        user.is_admin = True
        logging.info("    %r", user.username)
        not_found.discard(user.username)

    if not_found:
        for username in not_found:
            session.add(User(username=username, is_admin=True))
            logging.info("    %r (new user)", username)

    await session.commit()
    logging.info("Done.")


async def remove_admins(session: AsyncSession, usernames: list[str]) -> None:
    logging.info("Removing SUPERADMIN users:")
    result = await session.execute(select(User).where(User.username.in_(usernames)).order_by(User.username))
    users = result.scalars().all()

    not_found = set(usernames)
    for user in users:
        logging.info("    %r", user.username)
        user.is_admin = False
        not_found.discard(user.username)

    if not_found:
        logging.info("Not found:")
        for username in not_found:
            logging.info("    %r", username)

    await session.commit()
    logging.info("Done.")


async def list_admins(session: AsyncSession) -> None:
    result = await session.execute(select(User).filter_by(is_admin=True).order_by(User.username))
    admins = result.scalars().all()
    logging.info("Listing users with SUPERADMIN role:")
    for admin in admins:
        logging.info("    %r", admin.username)
    logging.info("Done.")


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Manage admin users.")
    subparsers = parser.add_subparsers(dest="command", required=True)

    parser_add = subparsers.add_parser("add", help="Add admin privileges to users")
    parser_add.add_argument("usernames", nargs="+", help="Usernames to add as admins")
    parser_add.set_defaults(func=add_admins)

    parser_remove = subparsers.add_parser("remove", help="Remove admin privileges from users")
    parser_remove.add_argument("usernames", nargs="+", help="Usernames to remove from admins")
    parser_remove.set_defaults(func=remove_admins)

    parser_list = subparsers.add_parser("list", help="List all admins")
    parser_list.set_defaults(func=list_admins)

    return parser


async def main(args: argparse.Namespace, session: AsyncSession) -> None:
    async with session:
        if args.command == "list":
            # 'list' command does not take additional arguments
            await args.func(session)
        else:
            await args.func(session, args.usernames)


if __name__ == "__main__":
    settings = Settings()
    if settings.server.logging.setup:
        setup_logging(settings.server.logging.get_log_config_path())

    engine = create_async_engine(settings.database.url)
    SessionLocal = async_sessionmaker(autocommit=False, autoflush=False, bind=engine, class_=AsyncSession)
    parser = create_parser()
    args = parser.parse_args()
    session = SessionLocal()
    asyncio.run(main(args, session))



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from horizon.commons.exceptions.base import ApplicationError


class BadRequestError(ApplicationError):
    """Bad request error.

    This exception should be raised when a request cannot be processed due to
    client-side errors (e.g., invalid data, duplicate entries).

    Examples
    --------

    >>> from horizon.commons.exceptions import BadRequestError
    >>> raise BadRequestError("Duplicate username detected. Each username must appear only once.")
    Traceback (most recent call last):
    horizon.commons.exceptions.BadRequestError: Duplicate username detected. Each username must appear only once.
    """

    reason: str
    """Bad request reason message"""

    def __init__(self, reason: str):
        self.reason = reason

    @property
    def message(self) -> str:
        return self.reason

    @property
    def details(self) -> dict:
        return {}



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime, timezone
from functools import partial

from sqlalchemy import DateTime, func
from sqlalchemy.orm import Mapped, mapped_column


class TimestampMixin:
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        server_default=func.now(),
        nullable=False,
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=partial(datetime.now, tz=timezone.utc),
        nullable=False,
    )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles

from horizon.backend.settings.server.static_files import StaticFilesSettings


def apply_static_files(app: FastAPI, settings: StaticFilesSettings) -> FastAPI:
    """Add static files serving middleware to the application."""
    if not settings.enabled:
        return app

    # https://fastapi.tiangolo.com/how-to/custom-docs-ui-assets/#serve-the-static-files
    app.mount("/static", StaticFiles(directory=settings.directory), name="static")
    return app



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path

from pydantic import BaseModel, Field, validator


class StaticFilesSettings(BaseModel):
    """Static files serving settings.

    Files are served at ``/static`` endpoint.

    Examples
    --------

    .. code-block:: bash

        HORIZON__SERVER__STATIC_FILES__ENABLED=True
        HORIZON__SERVER__STATIC_FILES__DIRECTORY=/app/horizon/backend/static
    """

    enabled: bool = Field(default=True, description="Set to ``True`` to enable static file serving")
    directory: Path = Field(
        default=Path("docs/_static"),
        description="Directory containing static files",
    )

    @validator("directory")
    def _validate_directory(cls, value: Path) -> Path:
        if not value.exists():
            raise ValueError(f"Directory '{value}' does not exist")
        if not value.is_dir():
            raise ValueError(f"Path '{value}' is not a directory")
        return value



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

import textwrap
from typing import Any, Dict, Optional

from pydantic import AnyHttpUrl, BaseModel, Field


class SwaggerSettings(BaseModel):
    """Swagger UI settings.

    SwaggerUI is served at ``/docs`` endpoint.

    Examples
    --------

    .. code-block:: bash

        HORIZON__SERVER__OPENAPI__SWAGGER__ENABLED=True
        HORIZON__SERVER__OPENAPI__SWAGGER__JS_URL=/static/swagger/swagger-ui-bundle.js
        HORIZON__SERVER__OPENAPI__SWAGGER__CSS_URL=/static/swagger/swagger-ui.css
    """

    enabled: bool = Field(default=True, description="Set to ``True`` to enable Swagger UI endpoint")
    js_url: str = Field(
        default="https://cdn.jsdelivr.net/npm/swagger-ui-dist@5/swagger-ui-bundle.js",
        description="URL for Swagger UI JS",
    )
    css_url: str = Field(
        default="https://cdn.jsdelivr.net/npm/swagger-ui-dist@5/swagger-ui.css",
        description="URL for Swagger UI CSS",
    )
    extra_parameters: Dict[str, Any] = Field(
        default_factory=dict,
        description=textwrap.dedent(
            """
            Additional parameters to pass to Swagger UI.
            See `FastAPI documentation <https://fastapi.tiangolo.com/how-to/configure-swagger-ui/>`_.
            """,
        ),
    )


class RedocSettings(BaseModel):
    """ReDoc settings.

    ReDOc is served at ``/redoc`` endpoint.

    Examples
    --------

    .. code-block:: bash

        HORIZON__SERVER__OPENAPI__REDOC__ENABLED=True
        HORIZON__SERVER__OPENAPI__REDOC__JS_URL=/static/redoc/redoc.standalone.js
    """

    enabled: bool = Field(default=True, description="Set to ``True`` to enable Redoc UI endpoint")
    js_url: str = Field(
        default="https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js",
        description="URL for Redoc UI JS, ``None`` to use default CDN URL",
    )


class LogoSettings(BaseModel):
    """OpenAPI's ``x-logo`` documentation settings.

    See `OpenAPI spec <https://redocly.com/docs/api-reference-docs/specification-extensions/x-logo/>`_
    for more details.

    Examples
    --------

    .. code-block:: bash

        HORIZON__SERVER__OPENAPI__LOGO__URL=/static/logo.svg
        HORIZON__SERVER__OPENAPI__LOGO__BACKGROUND_COLOR=ffffff
        HORIZON__SERVER__OPENAPI__LOGO__ALT_TEXT=Horizon logo
        HORIZON__SERVER__OPENAPI__LOGO__HREF=http://mycompany.domain.com
    """

    url: str = Field(
        default="/static/logo.svg",
        description="URL for application logo",
    )
    background_color: str = Field(
        default="ffffff",
        description="Background color in HEX RGB format, without ``#`` prefix",
    )
    alt_text: Optional[str] = Field(
        default="Horizon logo",
        description="Alternative text for ``<img>`` tag",
    )
    href: Optional[AnyHttpUrl] = Field(  # type: ignore[assignment]
        default="https://github.com/MobileTeleSystems/horizon",
        description="Clicking on logo will redirect to this URL",
    )


class FaviconSettings(BaseModel):
    """Favicon documentation settings.

    Examples
    --------

    .. code-block:: bash

        HORIZON__SERVER__OPENAPI__FAVICON__URL=/static/icon.svg
    """

    url: str = Field(
        default="/static/icon.svg",
        description="URL for application favicon",
    )


class OpenAPISettings(BaseModel):
    """OpenAPI Settings.

    OpenAPI.json is served at ``/openapi.json`` endpoint.

    Examples
    --------

    .. code-block:: bash

        HORIZON__SERVER__OPENAPI__ENABLED=True
        HORIZON__SERVER__OPENAPI__SWAGGER__ENABLED=True
        HORIZON__SERVER__OPENAPI__REDOC__ENABLED=True
    """

    enabled: bool = Field(default=True, description="Set to ``True`` to enable OpenAPI.json endpoint")
    swagger: SwaggerSettings = Field(
        default_factory=SwaggerSettings,
        description="Swagger UI settings",
    )
    redoc: RedocSettings = Field(
        default_factory=RedocSettings,
        description="ReDoc UI settings",
    )
    logo: LogoSettings = Field(
        default_factory=LogoSettings,
        description="Application logo settings",
    )
    favicon: FaviconSettings = Field(
        default_factory=FaviconSettings,
        description="Application favicon settings",
    )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from typing import Any

from horizon.commons.exceptions.base import ApplicationError


class AuthorizationError(ApplicationError):
    """Authorization request is failed.

    Examples
    --------

    >>> from horizon.commons.exceptions import AuthorizationError
    >>> raise AuthorizationError("User 'test' is disabled")
    Traceback (most recent call last):
    horizon.commons.exceptions.auth.AuthorizationError: User 'test' is disabled
    """

    def __init__(self, message: str, details: Any = None) -> None:
        self._message = message
        self._details = details

    @property
    def message(self) -> str:
        return self._message

    @property
    def details(self) -> Any:
        return self._details



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from fastapi import APIRouter

from horizon.backend.api.monitoring import router as monitoring_router
from horizon.backend.api.v1.router import router as v1_router

api_router = APIRouter()
api_router.include_router(monitoring_router)
api_router.include_router(v1_router)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from jose import ExpiredSignatureError, JWTError, jwt

from horizon.commons.exceptions import AuthorizationError


def sign_jwt(payload: dict, secret_key: str, security_algorithm: str) -> str:
    return jwt.encode(
        payload,
        secret_key,
        algorithm=security_algorithm,
    )


def decode_jwt(token: str, secret_key: str, security_algorithm: str) -> dict:
    try:
        result = jwt.decode(
            token,
            secret_key,
            algorithms=[security_algorithm],
        )
        if "exp" not in result:
            raise ExpiredSignatureError("Missing expiration time in token")

        return result
    except JWTError as e:
        raise AuthorizationError("Invalid token") from e



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
"""Add is_admin to User

Revision ID: ec64f7b42221
Revises: 4c60578f3f1d
Create Date: 2024-03-18 15:59:19.680251

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ec64f7b42221"
down_revision = "4c60578f3f1d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column("user", sa.Column("is_admin", sa.Boolean(), nullable=False, server_default=sa.false()))


def downgrade() -> None:
    op.drop_column("user", "is_admin")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from horizon.backend.db.models.base import Base
from horizon.backend.db.models.credentials_cache import CredentialsCache
from horizon.backend.db.models.hwm import HWM
from horizon.backend.db.models.hwm_history import HWMHistory
from horizon.backend.db.models.namespace import Namespace, NamespaceUserRoleInt
from horizon.backend.db.models.namespace_history import NamespaceHistory
from horizon.backend.db.models.namespace_user import NamespaceUser
from horizon.backend.db.models.user import User



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

from sqlalchemy import JSON, BigInteger, ForeignKey, String, Text, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship

from horizon.backend.db.mixins.changed_by import ChangedByMixin
from horizon.backend.db.models.base import Base
from horizon.backend.db.models.namespace import Namespace


class HWM(Base, ChangedByMixin):
    __tablename__ = "hwm"
    __table_args__ = (UniqueConstraint("namespace_id", "name", name="hwm_name_unique_per_namespace"),)

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)

    namespace_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("namespace.id", ondelete="CASCADE"),
        nullable=True,
        index=True,
    )
    namespace = relationship(Namespace, lazy="selectin")

    name: Mapped[str] = mapped_column(String(2048), nullable=False, index=True)  # noqa: WPS432
    description: Mapped[str] = mapped_column(Text(), nullable=False, default="")
    type: Mapped[str] = mapped_column(String(64), nullable=False)  # noqa: WPS432
    value: Mapped[str] = mapped_column(JSON(), nullable=False)
    entity: Mapped[Optional[str]] = mapped_column(Text(), nullable=True)
    expression: Mapped[Optional[str]] = mapped_column(Text(), nullable=True)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from sqlalchemy import BigInteger, ForeignKey, String
from sqlalchemy.orm import Mapped, mapped_column

from horizon.backend.db.models.base import Base


class NamespaceUser(Base):
    __tablename__ = "namespace_user"

    namespace_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("namespace.id", ondelete="CASCADE"),
        primary_key=True,
    )
    user_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("user.id", ondelete="CASCADE"),
        primary_key=True,
    )
    role: Mapped[str] = mapped_column(
        String(50),  # noqa: WPS432
        nullable=False,
    )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from pydantic import BaseModel, Field

try:
    from pydantic import ImportString
except ImportError:
    from pydantic import PyObject as ImportString  # type: ignore[no-redef] # noqa: WPS440

from horizon.backend.providers.auth.dummy import DummyAuthProvider


class AuthSettings(BaseModel):
    """Authorization-related settings.

    Here you can set auth provider class along with its options.

    Examples
    --------

    .. code-block:: bash

        # set settings.auth.provider = horizon.backend.providers.auth.dummy.DummyAuthProvider
        HORIZON__AUTH__PROVIDER=horizon.backend.providers.auth.dummy.DummyAuthProvider

        # pass access_key.secret_key = "secret" to DummyAuthProviderSettings
        HORIZON__AUTH__ACCESS_KEY__SECRET_KEY=secret
    """

    provider: ImportString = Field(  # type: ignore[assignment]
        default=DummyAuthProvider,
        description="Full name of auth provider class",
    )

    class Config:
        extra = "allow"



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from fastapi import FastAPI

from horizon.backend.db.models import User


class AuthProvider(ABC):
    """Basic class for all Auth providers.

    Constructor is called by FastAPI, and can use Dependency injection mechanism.
    See :obj:`~setup` for more details.
    """

    @classmethod
    @abstractmethod
    def setup(cls, app: FastAPI) -> FastAPI:
        """
        This method is called by :obj:`horizon.backend.application_factory`.

        Here you should add dependency overrides for auth provider,
        and return new ``app`` object.

        Examples
        --------

        .. code-block::

            from fastapi import FastAPI
            from my_awesome_auth_provider.settings import MyAwesomeAuthProviderSettings
            from horizon.backend.dependencies import Stub

            class MyAwesomeAuthProvider(AuthProvider):
                def setup(app):
                    app.dependency_overrides[AuthProvider] = MyAwesomeAuthProvider

                    # `settings_object_factory` returns MyAwesomeAuthProviderSettings object
                    app.dependency_overrides[MyAwesomeAuthProviderSettings] = settings_object_factory
                    return app

                def __init__(
                    self,
                    settings: Annotated[MyAwesomeAuthProviderSettings, Depends(Stub(MyAwesomeAuthProviderSettings))],
                ):
                    # settings object is set automatically by FastAPI's dependency_overrides
                    self.settings = settings
        """
        ...

    @abstractmethod
    async def get_current_user(self, access_token: str) -> User:
        """
        This method should return currently logged in user.

        Parameters
        ----------
        access_token : str
            JWT token got from ``Authorization: Bearer <token>`` header.

        Returns
        -------
        :obj:`horizon.backend.db.models.User`
            Current user object
        """
        ...

    @abstractmethod
    async def get_token(
        self,
        grant_type: Optional[str] = None,
        login: Optional[str] = None,
        password: Optional[str] = None,
        scopes: Optional[List[str]] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        This method should perform authentication and return JWT token.

        Parameters
        ----------
        See:
          * https://auth0.com/docs/get-started/authentication-and-authorization-flow/call-your-api-using-resource-owner-password-flow
          * https://connect2id.com/products/server/docs/api/token

        Returns
        -------
        Dict:
            .. code-block:: python

                {
                    "access_token": "some.jwt.token",
                    "token_type": "bearer",
                    "expires_in": 3600,
                }
        """
        ...



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
"""Add owner_id to namespace

Revision ID: 09be8fd79dbc
Revises: 2452f82ae06c
Create Date: 2024-02-29 11:22:25.802243

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "09be8fd79dbc"
down_revision = "2452f82ae06c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column("namespace", sa.Column("owner_id", sa.BigInteger(), nullable=True))
    op.execute(
        """
        UPDATE namespace
        SET owner_id = changed_by_user_id
        WHERE owner_id IS NULL
        """,
    )
    op.alter_column("namespace", "owner_id", nullable=False)
    op.create_foreign_key(
        op.f("fk__namespace__owner_id__user"),
        "namespace",
        "user",
        ["owner_id"],
        ["id"],
        ondelete="RESTRICT",
    )
    op.add_column("namespace_history", sa.Column("owner_id", sa.BigInteger(), nullable=True))
    op.create_foreign_key(
        op.f("fk__namespace_history__owner_id__user"),
        "namespace_history",
        "user",
        ["owner_id"],
        ["id"],
        ondelete="SET NULL",
    )
    op.execute(
        """
        INSERT INTO namespace_history (namespace_id, name, description, action, owner_id, changed_by_user_id)
        SELECT id, name, description, 'Set owner (automatic migration)', owner_id, NULL
        FROM namespace
        """,
    )


def downgrade() -> None:
    op.drop_constraint(op.f("fk__namespace_history__owner_id__user"), "namespace_history", type_="foreignkey")
    op.drop_column("namespace_history", "owner_id")
    op.drop_constraint(op.f("fk__namespace__owner_id__user"), "namespace", type_="foreignkey")
    op.drop_column("namespace", "owner_id")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from sqlalchemy.exc import IntegrityError

from horizon.backend.db.models import User
from horizon.backend.db.repositories.base import Repository
from horizon.commons.exceptions.entity import (
    EntityAlreadyExistsError,
    EntityNotFoundError,
)


class UserRepository(Repository[User]):
    async def count(self) -> int:
        return await self._count(where=[User.is_active.is_(True)])

    async def get_by_id(self, user_id: int) -> User:
        result = await self._get_by_id(user_id)
        if result is None:
            raise EntityNotFoundError("User", "id", user_id)
        return result

    async def get_by_username(self, username: str) -> User:
        user = await self._get(User.username == username)
        if not user:
            raise EntityNotFoundError("User", "username", username)
        return user

    async def create(
        self,
        username: str,
    ) -> User:
        try:
            result = await self._create(data={"username": username})
            await self._session.flush()
            return result
        except IntegrityError as e:
            raise EntityAlreadyExistsError("User", "username", username) from e

    async def get_or_create(
        self,
        username: str,
    ) -> User:
        result = await self._get(User.username == username)
        if not result:
            result = await self.create(username)
        return result



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

import textwrap
from pathlib import Path
from typing import Optional

from pydantic import BaseModel, Field
from typing_extensions import Literal

LOG_PATH = Path(__file__).parent.resolve()


class LoggingSettings(BaseModel):
    """Logging Settings.

    Examples
    --------

    Using ``json`` preset:

    .. code-block:: bash

        HORIZON__SERVER__LOGGING__SETUP=True
        HORIZON__SERVER__LOGGING__PRESET=json

    Passing custom logging config file:

    .. code-block:: bash

        HORIZON__SERVER__LOGGING__SETUP=True
        HORIZON__SERVER__LOGGING__CUSTOM_CONFIG_PATH=/some/logging.yml

    Setup logging in some other way, e.g. using `uvicorn args <https://www.uvicorn.org/settings/#logging>`_:

    .. code-block:: bash

        $ export HORIZON__SERVER__LOGGING__SETUP=False
        $ python -m horizon.backend --log-level debug
    """

    setup: bool = Field(
        default=True,
        description="If ``True``, setup logging during application start",
    )
    preset: Literal["json", "plain", "colored"] = Field(
        default="plain",
        description=textwrap.dedent(
            """
            Name of logging preset to use.

            There are few logging presets bundled to ``data-horizon[backend]`` package:

            .. dropdown:: ``plain`` preset

                This preset is recommended to use in environment which do not support colored output,
                e.g. CI jobs

                .. literalinclude:: ../../../horizon/backend/settings/server/log/plain.yml

            .. dropdown:: ``colored`` preset

                This preset is recommended to use in development environment,
                as it simplifies debugging. Each log record is output with color specific for a log level

                .. literalinclude:: ../../../horizon/backend/settings/server/log/colored.yml

            .. dropdown:: ``json`` preset

                This preset is recommended to use in production environment,
                as it allows to avoid writing complex log parsing configs. Each log record is output as JSON line

                .. literalinclude:: ../../../horizon/backend/settings/server/log/json.yml
            """,
        ),
    )

    custom_config_path: Optional[Path] = Field(
        default=None,
        description=textwrap.dedent(
            """
            Path to custom logging configuration file. If set, overrides :obj:`~preset` value.

            File content should be in YAML format and conform
            `logging.dictConfig <https://docs.python.org/3/library/logging.config.html#logging-config-dictschema>`_.
            """,
        ),
    )

    def get_log_config_path(self) -> Path:
        return self.custom_config_path or LOG_PATH / f"{self.preset}.yml"



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0


def slugify(text: str) -> str:
    """Convert ``Some value`` to ``some-value``.

    Used to convert ``FastAPI.title`` to short name which can be used as label for Prometheus metrics.
    """
    return text.lower().strip().replace(" ", "-")



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

from sqlalchemy import JSON, BigInteger, ForeignKey, String, Text
from sqlalchemy.orm import Mapped, mapped_column

from horizon.backend.db.mixins.changed_by import ChangedByMixin
from horizon.backend.db.models.base import Base


class HWMHistory(Base, ChangedByMixin):
    __tablename__ = "hwm_history"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)

    hwm_id: Mapped[int] = mapped_column(BigInteger, index=True, nullable=True)

    namespace_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("namespace.id", ondelete="CASCADE"),
        nullable=True,
    )

    name: Mapped[str] = mapped_column(String(2048), nullable=False)  # noqa: WPS432
    description: Mapped[str] = mapped_column(Text(), nullable=False)
    type: Mapped[str] = mapped_column(String(64), nullable=False)  # noqa: WPS432
    value: Mapped[str] = mapped_column(JSON(), nullable=False)
    entity: Mapped[Optional[str]] = mapped_column(Text(), nullable=True)
    expression: Mapped[Optional[str]] = mapped_column(Text(), nullable=True)
    action: Mapped[str] = mapped_column(String(255), nullable=False)  # noqa: WPS432



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
"""Drop is_deleted column from user

Revision ID: 2452f82ae06c
Revises: 5c7a5c5a193b
Create Date: 2024-02-27 14:55:54.744041

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2452f82ae06c"
down_revision = "5c7a5c5a193b"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute('DELETE FROM "user" WHERE is_deleted = TRUE')
    op.drop_column("user", "is_deleted")


def downgrade() -> None:
    op.add_column("user", sa.Column("is_deleted", sa.BOOLEAN(), autoincrement=False, nullable=False))



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from pydantic import BaseModel


class AuthTokenResponseV1(BaseModel):
    """Authorization response."""

    access_token: str
    token_type: str
    expires_at: float



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from typing import Generic, List, Optional, TypeVar

from pydantic import BaseModel, Field
from pydantic import __version__ as pydantic_version

if pydantic_version >= "2":
    from pydantic import BaseModel as GenericModel
else:
    from pydantic.generics import GenericModel  # type: ignore[no-redef] # noqa: WPS440

from horizon.commons.dto import Pagination


class PaginateQueryV1(BaseModel):
    """Basic class with pagination query params."""

    page: int = Field(gt=0, default=1, description="Page number")
    page_size: int = Field(gt=0, le=50, default=20, description="Number of items per page")  # noqa: WPS432


class PageMetaResponseV1(BaseModel):
    """Page metadata response."""

    page: int = Field(description="Page number")
    page_size: int = Field(description="Number of items per page")
    total_count: int = Field(description="Total number of items")
    pages_count: int = Field(description="Number of items returned in current page")
    has_next: bool = Field(description="Is there a next page")
    has_previous: bool = Field(description="Is there a next page")
    next_page: Optional[int] = Field(description="Next page number, if any")
    previous_page: Optional[int] = Field(description="Previous page number, if any")


T = TypeVar("T", bound=BaseModel)


class PageResponseV1(GenericModel, Generic[T]):
    """Page response."""

    meta: PageMetaResponseV1 = Field(description="Page metadata")
    items: List[T] = Field(description="Page content")

    @classmethod
    def from_pagination(cls, pagination: Pagination):
        return cls(
            meta=PageMetaResponseV1(
                page=pagination.page,
                pages_count=pagination.pages_count,
                page_size=pagination.page_size,
                total_count=pagination.total_count,
                has_next=pagination.has_next,
                has_previous=pagination.has_previous,
                next_page=pagination.next_page,
                previous_page=pagination.previous_page,
            ),
            items=pagination.items,
        )



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from typing import List, Optional, Set

from pydantic import BaseModel, Field, validator

from horizon.commons.schemas.v1 import NamespaceUserRole


class PermissionResponseItemV1(BaseModel):
    """Represents a single permission entry in a response, linking a user with their role within a namespace."""

    username: str
    role: NamespaceUserRole


class PermissionsResponseV1(BaseModel):
    """Wraps a list of permission entries for a namespace, returned by the GET endpoint."""

    permissions: List[PermissionResponseItemV1]


class PermissionUpdateRequestItemV1(BaseModel):
    """Represents a single permission entry in a request, specifying a desired role for a user within a namespace."""

    username: str
    role: Optional[NamespaceUserRole] = Field(
        default=None,
        description="The role to be assigned to the user within the namespace."
        " A value of `None` indicates that the permission should be removed.",
    )


class PermissionsUpdateRequestV1(BaseModel):
    """Wraps a list of permission modification requests for a namespace, used by the PATCH endpoint."""

    permissions: List[PermissionUpdateRequestItemV1] = Field(
        description="A list of modifications to the namespace's permissions."
        " Each entry specifies a user and the role they should have or be removed from.",
    )

    @validator("permissions")
    def _ensure_unique_usernames_and_single_owner(
        cls,
        permissions: List[PermissionUpdateRequestItemV1],
    ) -> List[PermissionUpdateRequestItemV1]:
        seen: Set[str] = set()
        owner_count = 0
        for perm in permissions:
            username, role = perm.username, perm.role

            if username in seen:
                raise ValueError(f"Duplicate username detected: {username}. Each username must appear only once.")
            seen.add(username)

            if role == NamespaceUserRole.OWNER:
                owner_count += 1

        if owner_count > 1:
            raise ValueError("Multiple owner role assignments detected. Only one owner can be assigned.")

        return permissions



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from fastapi import APIRouter

from horizon.commons.schemas import PingResponse

router = APIRouter(tags=["Monitoring"], prefix="/monitoring")


@router.get("/ping", summary="Check if server is alive")
async def ping() -> PingResponse:
    return PingResponse()



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

# _raw_version could contain pre-release version, like 0.0.1dev123
# value is updated automatically by `poetry version ...` and poetry-bumpversion plugin
_raw_version = "1.0.1"

# version always contain only release number like 0.0.1
__version__ = ".".join(_raw_version.split(".")[:3])  # noqa: WPS410

# version tuple always contains only integer parts, like (0, 0, 1)
__version_tuple__ = tuple(map(int, __version__.split(".")))



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import List, Optional, TypeVar, Union

from authlib.integrations.requests_client import OAuth2Session
from pydantic import BaseModel, Field, root_validator, validator
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from horizon import __version__ as horizon_version
from horizon.client.base import BaseClient
from horizon.commons.schemas import PingResponse
from horizon.commons.schemas.v1 import (
    HWMBulkCopyRequestV1,
    HWMBulkDeleteRequestV1,
    HWMCreateRequestV1,
    HWMHistoryPaginateQueryV1,
    HWMHistoryResponseV1,
    HWMListResponseV1,
    HWMPaginateQueryV1,
    HWMResponseV1,
    HWMUpdateRequestV1,
    NamespaceCreateRequestV1,
    NamespaceHistoryPaginateQueryV1,
    NamespaceHistoryResponseV1,
    NamespacePaginateQueryV1,
    NamespaceResponseV1,
    NamespaceUpdateRequestV1,
    PageResponseV1,
    PermissionsResponseV1,
    PermissionsUpdateRequestV1,
    UserResponseV1,
    UserResponseV1WithAdmin,
)

ResponseSchema = TypeVar("ResponseSchema", bound=BaseModel)


class RetryConfig(BaseModel):
    """
    Configuration for request retries in case of network errors or specific status codes.
    If provided, it customizes the retry behavior for requests made by the client.
    `urllib3 retry documentation <https://urllib3.readthedocs.io/en/stable/reference/urllib3.util.html>`_.

    Parameters
    ----------
    total : int, default: ``3``
        The maximum number of retry attempts to make.

    backoff_factor : float, default: ``0.1``
        A backoff factor to apply between attempts after the second try.

    status_forcelist : list[int], default: ``[502, 503, 504]``
        A set of HTTP status codes that we should force a retry on.

    backoff_jitter : float, default: ``None``
        A random jitter amount (between 0 and 1) to add to the backoff delay.
        Helps to avoid "thundering herd" issues by randomizing the delay
        times between retries.

        .. note::

            Requires ``urllib>2.0``
    """

    total: int = 3
    backoff_factor: float = 0.1
    status_forcelist: List[int] = [502, 503, 504]
    backoff_jitter: Optional[float] = None


class TimeoutConfig(BaseModel):
    """
    Configuration for connection and request timeouts.
    If provided, it customizes the timeout behavior for requests made by the client.
    `requests timeout documentation <https://requests.readthedocs.io/en/latest/user/advanced/#timeouts>`_.

    Parameters
    ----------
    connection_timeout : float, default: ``3``
        The maximum number of seconds to wait for a connection to the server.

    request_timeout : float, default: ``5``
        The maximum number of seconds to wait for a response from the server.
    """

    connection_timeout: float = 3
    request_timeout: float = 5


class HorizonClientSync(BaseClient[OAuth2Session]):
    """Sync Horizon client implementation, based on ``authlib`` and ``requests``.

    Parameters
    ----------

    base_url : str
        URL of Horizon API, e.g. ``https://some.domain.com/api``

    auth : :obj:`BaseAuth <horizon.client.auth.base.BaseAuth>`
        Authentication class

    retry : :obj:`RetryConfig <horizon.client.sync.RetryConfig>`
        Configuration for request retries.

    timeout : :obj:`TimeoutConfig <horizon.client.sync.TimeoutConfig>`
        Configuration for request timeouts.

    session : :obj:`authlib.integrations.requests_client.OAuth2Session`
        Custom session object. Inherited from :obj:`requests.Session`, so you can pass custom
        session options.

    Examples
    --------

    Using default parameters:

    >>> from horizon.client.auth import LoginPassword
    >>> from horizon.client.sync import HorizonClientSync
    >>> client = HorizonClientSync(
    ...     base_url="https://some.domain.com/api",
    ...     auth=LoginPassword(login="me", password="12345"),
    ... )

    Customize retry and timeout:

    >>> from horizon.client.auth import LoginPassword
    >>> from horizon.client.sync import HorizonClientSync, RetryConfig, TimeoutConfig
    >>> client = HorizonClientSync(
    ...     base_url="https://some.domain.com/api",
    ...     auth=LoginPassword(login="me", password="12345"),
    ...     retry=RetryConfig(total=2, backoff_factor=10, status_forcelist=[500, 503]),
    ...     timeout=TimeoutConfig(request_timeout=3.5),
    ... )
    """

    retry: RetryConfig = Field(default_factory=RetryConfig)
    timeout: TimeoutConfig = Field(default_factory=TimeoutConfig)

    def authorize(self) -> None:
        """Fetch and set access token (if required).

        Raises
        ------
        :obj:`horizon.commons.exceptions.AuthorizationError`
            Authorization failed

        Examples
        --------

        >>> client.authorize()
        """

        session: OAuth2Session = self.session  # type: ignore[assignment]
        token_kwargs = self.auth.fetch_token_kwargs(self.base_url)
        if token_kwargs:
            session.token = session.fetch_token(**token_kwargs)

        # token will not be verified until we call any endpoint
        # do not call ``self.whoami`` here to avoid recursion
        timeout = (self.timeout.connection_timeout, self.timeout.request_timeout)
        response = session.request("GET", f"{self.base_url}/v1/users/me", timeout=timeout)
        self._handle_response(response, UserResponseV1)

    def close(self) -> None:
        """Close session.

        Examples
        --------

        >>> client.close()
        """
        session: OAuth2Session = self.session  # type: ignore[assignment]
        session.close()

    def __enter__(self):
        """Enter session as context manager. Similar to :obj:`requests.Session` behavior.

        Exiting context manager closes opened session.

        Examples
        --------

        >>> with client:
        ...    ...
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def ping(self) -> PingResponse:
        """Ping Horizon server.

        Examples
        --------

        >>> client.ping()
        PingResponse(status="ok")
        """
        return self._request(  # type: ignore[return-value]
            "GET",
            f"{self.base_url}/monitoring/ping",
            response_class=PingResponse,
        )

    def whoami(self) -> Union[UserResponseV1WithAdmin, UserResponseV1]:
        """Get current user info.

        Examples
        --------

        >>> client.whoami()
        UserResponseV1(
            id=1,
            username="me",
        )

        >>> client.whoami()  # for a superadmin user:
        UserResponseV1WithAdmin(
            id=1,
            username="admin",
            is_admin=True,
        )

        """
        return self._request(  # type: ignore[return-value]
            "GET",
            f"{self.base_url}/v1/users/me",
            response_class=Union[UserResponseV1WithAdmin, UserResponseV1],  # type: ignore[arg-type]
        )

    def paginate_namespaces(
        self,
        query: NamespacePaginateQueryV1 | None = None,
    ) -> PageResponseV1[NamespaceResponseV1]:
        """Get page with namespaces.

        Parameters
        ----------
        query : :obj:`NamespacePaginateQueryV1 <horizon.commons.schemas.v1.namespace.NamespacePaginateQueryV1>`
            Namespace query parameters

        Returns
        -------
        :obj:`PageResponseV1 <horizon.commons.schemas.v1.pagination.PageResponseV1>` of :obj:`NamespaceResponseV1 <horizon.commons.schemas.v1.namespace.NamespaceResponseV1>`
            List of namespaces, limited and filtered by query parameters.

        Examples
        --------

        Get all namespaces:

        >>> client.paginate_namespaces()
        PageResponseV1[NamespaceResponseV1](
            meta=PageMetaResponseV1(
                page=1,
                pages_count=1,
                total_count=10,
                page_size=20,
                has_next=False,
                has_previous=False,
                next_page=None,
                previous_page=None,
            ),
            items=[NamespaceResponseV1(...), ...],
        )

        Get all namespaces starting with a page number and page size:

        >>> from horizon.commons.schemas.v1 import NamespacePaginateQueryV1
        >>> namespace_query = NamespacePaginateQueryV1(page=2, page_size=20)
        >>> client.paginate_namespaces(query=namespace_query)
        PageResponseV1[NamespaceResponseV1](
            meta=PageMetaResponseV1(
                page=2,
                pages_count=3,
                total_count=50,
                page_size=20,
                has_next=True,
                has_previous=True,
                next_page=3,
                previous_page=1,
            ),
            items=[NamespaceResponseV1(...), ...],
        )

        Search for namespace with specific name:

        >>> from horizon.commons.schemas.v1 import NamespacePaginateQueryV1
        >>> namespace_query = NamespacePaginateQueryV1(name="my_namespace")
        >>> client.paginate_namespaces(query=namespace_query)
        PageResponseV1[NamespaceResponseV1](
            meta=PageMetaResponseV1(
                page=1,
                pages_count=1,
                total_count=1,
                page_size=10,
                has_next=False,
                has_previous=False,
                next_page=None,
                previous_page=None,
            ),
            items=[
                NamespaceResponseV1(name="my_namespace", ...),
            ],
        )
        """
        query = query or NamespacePaginateQueryV1()
        return self._request(  # type: ignore[return-value]
            "GET",
            f"{self.base_url}/v1/namespaces/",
            response_class=PageResponseV1[NamespaceResponseV1],
            params=query.dict(),
        )

    def get_namespace(self, namespace_id: int) -> NamespaceResponseV1:
        """Get namespace by name.

        Parameters
        ----------
        namespace_id : int
            Namespace name to get

        Returns
        -------
        :obj:`NamespaceResponseV1 <horizon.commons.schemas.v1.namespace.NamespaceResponseV1>`
            Namespace

        Raises
        ------
        :obj:`EntityNotFoundError <horizon.commons.exceptions.entity.EntityNotFoundError>`
            Namespace not found

        Examples
        --------

        >>> client.get_namespace(namespace_id=123)
        NamespaceResponseV1(
            id=123,
            name="my_namespace",
            ...
        )
        """
        return self._request(  # type: ignore[return-value]
            "GET",
            f"{self.base_url}/v1/namespaces/{namespace_id}",
            response_class=NamespaceResponseV1,
        )

    def create_namespace(self, data: NamespaceCreateRequestV1) -> NamespaceResponseV1:
        """Create new namespace.

        Parameters
        ----------
        namespace : :obj:`NamespaceCreateRequestV1 <horizon.commons.schemas.v1.namespace.NamespaceCreateRequestV1>`
            Namespace to create

        Returns
        -------
        :obj:`NamespaceResponseV1 <horizon.commons.schemas.v1.namespace.NamespaceResponseV1>`
            Created namespace

        Raises
        ------
        :obj:`EntityAlreadyExistsError <horizon.commons.exceptions.entity.EntityAlreadyExistsError>`
            Namespace with the same name already exists

        Examples
        --------

        >>> from horizon.commons.schemas.v1 import NamespaceCreateRequestV1
        >>> to_create = NamespaceCreateRequestV1(name="my_namespace")
        >>> client.create_namespace(data=to_create)
        NamespaceResponseV1(
            id=123,
            name="my_namespace",
            ...
        )
        """
        return self._request(  # type: ignore[return-value]
            "POST",
            f"{self.base_url}/v1/namespaces/",
            json=data.dict(),
            response_class=NamespaceResponseV1,
        )

    def update_namespace(self, namespace_id: int, changes: NamespaceUpdateRequestV1) -> NamespaceResponseV1:
        """Update existing namespace.

        Parameters
        ----------
        namespace_id : int
            Namespace name to update

        changes : :obj:`NamespaceUpdateRequestV1 <horizon.commons.schemas.v1.namespace.NamespaceUpdateRequestV1>`
            Changes to namespace object

        Returns
        -------
        :obj:`NamespaceResponseV1 <horizon.commons.schemas.v1.namespace.NamespaceResponseV1>`
            Updated namespace

        Raises
        ------
        :obj:`EntityNotFoundError <horizon.commons.exceptions.entity.EntityNotFoundError>`
            Namespace not found
        :obj:`EntityAlreadyExistsError <horizon.commons.exceptions.entity.EntityAlreadyExistsError>`
            Namespace with the same name already exists
        :obj:`PermissionDeniedError <horizon.commons.exceptions.permission.PermissionDeniedError>`
            Permission denied for performing the requested action.

        Examples
        --------

        >>> from horizon.commons.schemas.v1 import NamespaceUpdateRequestV1
        >>> to_update = NamespaceUpdateRequestV1(name="new_namespace_name")
        >>> client.update_namespace(namespace_id=123, changes=to_update)
        NamespaceResponseV1(
            id=123,
            name="new_namespace_name",
            ...
        )
        """
        return self._request(  # type: ignore[return-value]
            "PATCH",
            f"{self.base_url}/v1/namespaces/{namespace_id}",
            json=changes.dict(exclude_unset=True),
            response_class=NamespaceResponseV1,
        )

    def delete_namespace(self, namespace_id: int) -> None:
        """Delete existing namespace.

        Parameters
        ----------
        namespace_id : int
            Namespace name to delete

        Raises
        ------
        :obj:`EntityNotFoundError <horizon.commons.exceptions.entity.EntityNotFoundError>`
            Namespace not found
        :obj:`PermissionDeniedError <horizon.commons.exceptions.permission.PermissionDeniedError>`
            Permission denied for performing the requested action.

        Examples
        --------

        >>> client.delete_namespace(namespace_id=123)
        """
        self._request(
            "DELETE",
            f"{self.base_url}/v1/namespaces/{namespace_id}",
        )

    def paginate_namespace_history(
        self,
        query: NamespaceHistoryPaginateQueryV1,
    ) -> PageResponseV1[NamespaceHistoryResponseV1]:
        """Get page with namespace changes history.

        Parameters
        ----------
        query : :obj:`NamespaceHistoryPaginateQueryV1 <horizon.commons.schemas.v1.namespace_history.NamespaceHistoryPaginateQueryV1>`
            Namespace history query parameters

        Returns
        -------
        :obj:`PageResponseV1 <horizon.commons.schemas.v1.pagination.PageResponseV1>` of :obj:`NamespaceHistoryResponseV1 <horizon.commons.schemas.v1.namespace_history.NamespaceHistoryResponseV1>`
            List of namespace history items, limited and filtered by query parameters.

        Examples
        --------

        Get all changes of specific namespace:

        >>> from horizon.commons.schemas.v1 import NamespacePaginateQueryV1
        >>> namespace_query = NamespacePaginateQueryV1(namespace_id=234)
        >>> client.paginate_namespace(query=namespace_query)
        PageResponseV1[NamespaceHistoryResponseV1](
            meta=PageMetaResponseV1(
                page=1,
                pages_count=2,
                total_count=10,
                page_size=10,
                has_next=True,
                has_previous=False,
                next_page=2,
                previous_page=None,
            ),
            items=[NamespaceHistoryResponseV1(namespace_id=234, ...), ...],
        )

        Get all changes of specific namespace starting with a page number and page size:

        >>> from horizon.commons.schemas.v1 import NamespacePaginateQueryV1
        >>> namespace_query = NamespacePaginateQueryV1(namespace_id=234, page=2, page_size=20)
        >>> client.paginate_namespace(query=namespace_query)
        PageResponseV1[NamespaceHistoryResponseV1](
            meta=PageMetaResponseV1(
                page=2,
                pages_count=3,
                total_count=50,
                page_size=20,
                has_next=True,
                has_previous=True,
                next_page=3,
                previous_page=1,
            ),
            items=[NamespaceHistoryResponseV1(namespace_id=234, ...), ...],
        )
        """
        return self._request(  # type: ignore[return-value]
            "GET",
            f"{self.base_url}/v1/namespace-history/",
            response_class=PageResponseV1[NamespaceHistoryResponseV1],
            params=query.dict(exclude_unset=True),
        )

    def paginate_hwm(
        self,
        query: HWMPaginateQueryV1,
    ) -> PageResponseV1[HWMResponseV1]:
        """Get page with HWMs.

        Parameters
        ----------
        query : :obj:`HWMPaginateQueryV1 <horizon.commons.schemas.v1.hwm.HWMPaginateQueryV1>`
            HWM query parameters

        Returns
        -------
        :obj:`PageResponseV1 <horizon.commons.schemas.v1.pagination.PageResponseV1>` of :obj:`HWMResponseV1 <horizon.commons.schemas.v1.hwm.HWMResponseV1>`
            List of HWM, limited and filtered by query parameters.

        Examples
        --------

        Get all HWM in namespace with specific id:

        >>> from horizon.commons.schemas.v1 import HWMPaginateQueryV1
        >>> hwm_query = HWMPaginateQueryV1(namespace_id=123)
        >>> client.paginate_hwm(query=hwm_query)
        PageResponseV1[HWMResponseV1](
            meta=PageMetaResponseV1(
                page=1,
                pages_count=2,
                total_count=10,
                page_size=10,
                has_next=True,
                has_previous=False,
                next_page=2,
                previous_page=None,
            ),
            items=[HWMResponseV1(namespace_id=123, ...), ...],
        )

        Get all HWM in namespace starting with a page number and page size:

        >>> from horizon.commons.schemas.v1 import HWMPaginateQueryV1
        >>> hwm_query = HWMPaginateQueryV1(namespace_id=123, page=2, page_size=20)
        >>> client.paginate_hwm(query=hwm_query)
        PageResponseV1[HWMResponseV1](
            meta=PageMetaResponseV1(
                page=2,
                pages_count=3,
                total_count=50,
                page_size=20,
                has_next=True,
                has_previous=True,
                next_page=3,
                previous_page=1,
            ),
            items=[HWMResponseV1(namespace_id=123, ...), ...],
        )

        Search for HWM with specific namespace and name:

        >>> from horizon.commons.schemas.v1 import HWMPaginateQueryV1
        >>> hwm_query = HWMPaginateQueryV1(namespace_id=123, name="my_hwm")
        >>> client.paginate_hwm(query=hwm_query)
        PageResponseV1[HWMResponseV1](
            meta=PageMetaResponseV1(
                page=1,
                pages_count=1,
                total_count=1,
                page_size=10,
                has_next=False,
                has_previous=False,
                next_page=None,
                previous_page=None,
            ),
            items=[
                HWMResponseV1(namespace_id=123, name="my_hwm", ...),
            ],
        )
        """
        return self._request(  # type: ignore[return-value]
            "GET",
            f"{self.base_url}/v1/hwm/",
            response_class=PageResponseV1[HWMResponseV1],
            params=query.dict(exclude_unset=True),
        )

    def get_hwm(self, hwm_id: int) -> HWMResponseV1:
        """Get HWM.

        Parameters
        ----------
        hwm_id : int
            HWM id to get

        Returns
        -------
        :obj:`HWMResponseV1 <horizon.commons.schemas.v1.hwm.HWMResponseV1>`
            HWM

        Raises
        ------
        :obj:`EntityNotFoundError <horizon.commons.exceptions.entity.EntityNotFoundError>`
            HWM not found

        Examples
        --------

        >>> client.get_hwm(hwm_id=234)
        HWMResponseV1(
            id=234,
            namespace_id=123,
            name="my_hwm",
            ...
        )
        """
        return self._request(  # type: ignore[return-value]
            "GET",
            f"{self.base_url}/v1/hwm/{hwm_id}",
            response_class=HWMResponseV1,
        )

    def create_hwm(self, data: HWMCreateRequestV1) -> HWMResponseV1:
        """Create new HWM.

        Parameters
        ----------
        data : :obj:`HWMCreateRequestV1 <horizon.commons.schemas.v1.hwm.HWMCreateRequestV1>`
            HWM data

        Returns
        -------
        :obj:`HWMResponseV1 <horizon.commons.schemas.v1.hwm.HWMResponseV1>`
            Created HWM

        Raises
        ------
        :obj:`EntityNotFoundError <horizon.commons.exceptions.entity.EntityNotFoundError>`
            Namespace not found
        :obj:`EntityAlreadyExistsError <horizon.commons.exceptions.entity.EntityAlreadyExistsError>`
            HWM with the same name already exists
        :obj:`PermissionDeniedError <horizon.commons.exceptions.permission.PermissionDeniedError>`
            Permission denied for performing the requested action.

        Examples
        --------

        >>> from horizon.commons.schemas.v1 import HWMCreateRequestV1
        >>> to_create = HWMCreateRequestV1(
        ...     namespace_id=123,
        ...     name="my_hwm",
        ...     type="column_int",
        ...     value=5678,
        ... )
        >>> client.create_hwm(data=to_create)
        HWMResponseV1(
            namespace_id=123,
            id=234,
            name="my_hwm",
            type="column_int",
            value=5678,
            ...,
        )
        """
        return self._request(  # type: ignore[return-value]
            "POST",
            f"{self.base_url}/v1/hwm/",
            json=data.dict(exclude_unset=True),
            response_class=HWMResponseV1,
        )

    def update_hwm(self, hwm_id: int, changes: HWMUpdateRequestV1) -> HWMResponseV1:
        """Update existing HWM.

        Parameters
        ----------
        hwm_id : int
            HWM id to update
        changes : :obj:`HWMUpdateRequestV1 <horizon.commons.schemas.v1.hwm.HWMUpdateRequestV1>`
            HWM changes

        Returns
        -------
        :obj:`HWMResponseV1 <horizon.commons.schemas.v1.hwm.HWMResponseV1>`
            Updated HWM

        Raises
        ------
        :obj:`EntityNotFoundError <horizon.commons.exceptions.entity.EntityNotFoundError>`
            HWM not found
        :obj:`EntityAlreadyExistsError <horizon.commons.exceptions.entity.EntityAlreadyExistsError>`
            HWM with the same name already exists
        :obj:`PermissionDeniedError <horizon.commons.exceptions.permission.PermissionDeniedError>`
            Permission denied for performing the requested action.

        Examples
        --------

        >>> from horizon.commons.schemas.v1 import HWMUpdateRequestV1
        >>> to_update = HWMUpdateRequestV1(type="column_int", value=5678)
        >>> client.update_hwm(hwm_id=234, changes=to_update)
        HWMResponseV1(
            namespace_id=123,
            id=234,
            name="my_hwm",
            type="column_int",
            value=5678,
            ...,
        )
        """
        return self._request(  # type: ignore[return-value]
            "PATCH",
            f"{self.base_url}/v1/hwm/{hwm_id}",
            json=changes.dict(exclude_unset=True),
            response_class=HWMResponseV1,
        )

    def delete_hwm(self, hwm_id: int) -> None:
        """Delete existing HWM.

        Parameters
        ----------
        hwm_id : int
            HWM id to delete

        Raises
        ------
        :obj:`EntityNotFoundError <horizon.commons.exceptions.entity.EntityNotFoundError>`
            HWM not found
        :obj:`PermissionDeniedError <horizon.commons.exceptions.permission.PermissionDeniedError>`
            Permission denied for performing the requested action.

        Examples
        --------

        >>> client.delete_hwm(hwm_id=234)
        """
        self._request(
            "DELETE",
            f"{self.base_url}/v1/hwm/{hwm_id}",
        )

    def bulk_copy_hwm(self, data: HWMBulkCopyRequestV1) -> HWMListResponseV1:
        """Copy HWMs from one namespace to another.

        .. note::

            Method ignores HWMs that are not related to provided source namespace, or does not exist.

        Parameters
        ----------
        data : :obj:`HWMBulkCopyRequestV1 <horizon.commons.schemas.v1.hwm.HWMBulkCopyRequestV1>`
            HWM copy data

        Returns
        -------
        :obj:`HWMListResponseV1 <horizon.commons.schemas.v1.hwm.HWMListResponseV1>`
            Copied HWMs

        Raises
        ------
        :obj:`EntityNotFoundError <horizon.commons.exceptions.entity.EntityNotFoundError>`
            Raised if any of the specified namespaces not found.
        :obj:`PermissionDeniedError <horizon.commons.exceptions.permission.PermissionDeniedError>`
            Permission denied for performing the requested action.

        Examples
        --------
        >>> from horizon.commons.schemas.v1 import HWMBulkCopyRequestV1
        >>> copy_data = HWMBulkCopyRequestV1(
        ...     source_namespace_id=123,
        ...     target_namespace_id=456,
        ...     hwm_ids=[1, 2, 3],
        ...     with_history=True,
        ... )
        >>> copied_hwms = client.bulk_copy_hwm(data=copy_data)
        [HWMResponseV1(...), HWMResponseV1(...), HWMResponseV1(...)]
        """
        return self._request(  # type: ignore[return-value]
            "POST",
            f"{self.base_url}/v1/hwm/copy",
            json=data.dict(),
            response_class=HWMListResponseV1,
        )

    def bulk_delete_hwm(self, namespace_id: int, hwm_ids: List[int]) -> None:
        """Bulk delete HWMs.

        .. note::

            Method ignores HWMs that are not related to provided namespace.

        Parameters
        ----------
        namespace_id : int
            Namespace ID where the HWMs belong.
        hwm_ids : List[int]
            List of HWM IDs to be deleted.

        Raises
        ------
        :obj:`PermissionDeniedError <horizon.commons.exceptions.permission.PermissionDeniedError>`
            Permission denied for performing the requested action.

        Examples
        --------

        >>> client.bulk_delete_hwm(
        ...     namespace_id=123,
        ...     hwm_ids=[234, 345, 456]
        ... )
        """

        data = HWMBulkDeleteRequestV1(namespace_id=namespace_id, hwm_ids=hwm_ids)
        self._request(
            "DELETE",
            f"{self.base_url}/v1/hwm/",
            json=data.dict(),
        )

    def get_namespace_permissions(self, namespace_id: int) -> PermissionsResponseV1:
        """Get permissions for a namespace.

        Parameters
        ----------
        namespace_id : int
            The ID of the namespace to get permissions for.

        Returns
        -------
        :obj:`PermissionsResponseV1 <horizon.commons.schemas.v1.permission.PermissionsResponseV1>`
            The permissions of the namespace.

        Raises
        ------
        :obj:`EntityNotFoundError <horizon.commons.exceptions.entity.EntityNotFoundError>`
            Namespace or provided user not found.
        :obj:`PermissionDeniedError <horizon.commons.exceptions.permission.PermissionDeniedError>`
            Permission denied for performing the requested action.

        Examples
        --------

        >>> client.get_namespace_permissions(namespace_id=234)
        """

        return self._request(  # type: ignore[return-value]
            "GET",
            f"{self.base_url}/v1/namespaces/{namespace_id}/permissions",
            response_class=PermissionsResponseV1,
        )

    def update_namespace_permissions(
        self,
        namespace_id: int,
        changes: PermissionsUpdateRequestV1,
    ) -> PermissionsResponseV1:
        """Update permissions for a namespace.

        Parameters
        ----------
        namespace_id : int
            The ID of the namespace to update permissions for.
        changes : :obj:`PermissionsUpdateRequestV1 <horizon.commons.schemas.v1.permission.PermissionsUpdateRequestV1>`
            The changes to apply to the namespace's permissions.

        Returns
        -------
        :obj:`PermissionsResponseV1 <horizon.commons.schemas.v1.permission.PermissionsResponseV1>`
            Actual permissions of the namespace.

        Raises
        ------
        :obj:`EntityNotFoundError <horizon.commons.exceptions.entity.EntityNotFoundError>`
            Namespace or provided user not found.
        :obj:`PermissionDeniedError <horizon.commons.exceptions.permission.PermissionDeniedError>`
            Permission denied for performing the requested action.
        :obj:`BadRequestError <horizon.commons.exceptions.bad_request.BadRequestError>`
            Bad request with incorrect operating logic.

        Examples
        --------

        >>> from horizon.commons.schemas.v1 import PermissionsUpdateRequestV1, PermissionUpdateRequestItemV1
        >>> to_update = PermissionsUpdateRequestV1([
        ...     PermissionUpdateRequestItemV1(username="new_owner", role="OWNER"),
        ...     PermissionUpdateRequestItemV1(username="add_developer", role="DEVELOPER"),
        ...     PermissionUpdateRequestItemV1(username="make_read_only", role=None),
        ... ])
        >>> client.update_namespace_permissions(namespace_id=234, changes=to_update)
        """
        return self._request(  # type: ignore[return-value]
            "PATCH",
            f"{self.base_url}/v1/namespaces/{namespace_id}/permissions",
            json=changes.dict(exclude_unset=True),
            response_class=PermissionsResponseV1,
        )

    def paginate_hwm_history(
        self,
        query: HWMHistoryPaginateQueryV1,
    ) -> PageResponseV1[HWMHistoryResponseV1]:
        """Get page with HWM changes history.

        Parameters
        ----------
        query : :obj:`HWMHistoryPaginateQueryV1 <horizon.commons.schemas.v1.hwm_history.HWMHistoryPaginateQueryV1>`
            HWM history query parameters

        Returns
        -------
        :obj:`PageResponseV1 <horizon.commons.schemas.v1.pagination.PageResponseV1>` of :obj:`HWMHistoryResponseV1 <horizon.commons.schemas.v1.hwm_history.HWMHistoryResponseV1>`
            List of HWM history items, limited and filtered by query parameters.

        Examples
        --------

        Get all changes of specific HWM:

        >>> from horizon.commons.schemas.v1 import HWMPaginateQueryV1
        >>> hwm_query = HWMPaginateQueryV1(hwm_id=234)
        >>> client.paginate_hwm(query=hwm_query)
        PageResponseV1[HWMHistoryResponseV1](
            meta=PageMetaResponseV1(
                page=1,
                pages_count=2,
                total_count=10,
                page_size=10,
                has_next=True,
                has_previous=False,
                next_page=2,
                previous_page=None,
            ),
            items=[HWMHistoryResponseV1(hwm_id=234, ...), ...],
        )

        Get all changes of specific HWM starting with a page number and page size:

        >>> from horizon.commons.schemas.v1 import HWMPaginateQueryV1
        >>> hwm_query = HWMPaginateQueryV1(hwm_id=234, page=2, page_size=20)
        >>> client.paginate_hwm(query=hwm_query)
        PageResponseV1[HWMHistoryResponseV1](
            meta=PageMetaResponseV1(
                page=2,
                pages_count=3,
                total_count=50,
                page_size=20,
                has_next=True,
                has_previous=True,
                next_page=3,
                previous_page=1,
            ),
            items=[HWMHistoryResponseV1(hwm_id=234, ...), ...],
        )
        """
        return self._request(  # type: ignore[return-value]
            "GET",
            f"{self.base_url}/v1/hwm-history/",
            response_class=PageResponseV1[HWMHistoryResponseV1],
            params=query.dict(exclude_unset=True),
        )

    # retry validator is called after "session" validators, when session already created by default or passed directly
    @root_validator(pre=False, skip_on_failure=True)
    def _configure_retries(cls, values):
        session = values.get("session")
        retry_config = values.get("retry")

        optional_retry_args = {}
        if retry_config.backoff_jitter is not None:
            # added to Retry class only in urllib3 2.0+
            optional_retry_args["backoff_jitter"] = retry_config.backoff_jitter

        retries = Retry(
            total=retry_config.total,
            backoff_factor=retry_config.backoff_factor,
            status_forcelist=retry_config.status_forcelist,
            allowed_methods=frozenset(("GET", "POST", "PUT", "PATCH", "DELETE")),
            **optional_retry_args,
        )
        adapter = HTTPAdapter(max_retries=retries)
        session.mount("https://", adapter)
        session.mount("http://", adapter)

        return values

    @validator("session", always=True)
    def _set_client_info(cls, session: OAuth2Session):
        session.headers["X-Client-Name"] = "python-horizon[sync]"
        session.headers["X-Client-Version"] = horizon_version
        return session

    def _request(
        self,
        method: str,
        url: str,
        response_class: type[ResponseSchema] | None = None,
        json: dict | None = None,
        params: dict | None = None,
    ) -> ResponseSchema | None:
        """Send request to backend and return ``response_class``, ``None`` or raise an exception."""

        session: OAuth2Session = self.session  # type: ignore[assignment]
        if not session.token or session.token.is_expired():
            self.authorize()

        timeout = (self.timeout.connection_timeout, self.timeout.request_timeout)
        response = session.request(method, url, json=json, params=params, timeout=timeout)
        return self._handle_response(response, response_class)



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import Any, List, Optional, Union

from pydantic import BaseModel, Field
from pydantic import __version__ as pydantic_version
from pydantic import root_validator, validator

from horizon.commons.dto import Unset
from horizon.commons.schemas.v1.pagination import PaginateQueryV1

MAX_NAME_LENGTH = 2048
MAX_TYPE_LENGTH = 64


class HWMResponseV1(BaseModel):
    """HWM response."""

    id: int = Field(description="HWM id")
    namespace_id: int = Field(description="Namespace id HWM is bound to")
    name: str = Field(description="HWM name, unique in the namespace")
    description: str = Field(description="HWM description")
    type: str = Field(description="HWM type, any non-empty string")
    value: Any = Field(description="HWM value, any JSON serializable value")
    entity: Optional[str] = Field(default=None, description="Name of entity associated with the HWM. Can be any string")
    expression: Optional[str] = Field(
        default=None,
        description="Expression used to calculate HWM value. Can be any string",
    )
    changed_at: datetime = Field(description="Timestamp of last change of the HWM data")
    changed_by: Optional[str] = Field(default=None, description="Latest user who changed the HWM data")

    class Config:
        if pydantic_version >= "2":
            from_attributes = True
        else:
            orm_mode = True


class HWMListResponseV1(BaseModel):
    hwms: List[HWMResponseV1]


class HWMPaginateQueryV1(PaginateQueryV1):
    """Query params for HWM pagination request."""

    namespace_id: int
    name: Optional[str] = Field(default=None, min_length=1, max_length=MAX_NAME_LENGTH)

    # more arguments can be added in future


class HWMCreateRequestV1(BaseModel):
    """Request body for HWM create request."""

    namespace_id: int
    name: str = Field(min_length=1, max_length=MAX_NAME_LENGTH)
    description: str = ""
    type: str = Field(min_length=1, max_length=MAX_TYPE_LENGTH)
    value: Any
    entity: Optional[str] = None
    expression: Optional[str] = None


class HWMUpdateRequestV1(BaseModel):
    """Request body for HWM update request.

    If field value is not set, it will not be updated.
    """

    name: Union[str, Unset] = Field(default=Unset(), min_length=1, max_length=MAX_NAME_LENGTH)
    description: Union[str, Unset] = Unset()
    type: Union[str, Unset] = Field(default=Unset(), min_length=1, max_length=MAX_TYPE_LENGTH)
    value: Union[Any, Unset] = Unset()
    entity: Union[str, None, Unset] = Unset()
    expression: Union[str, None, Unset] = Unset()

    class Config:
        arbitrary_types_allowed = True

    @root_validator(skip_on_failure=True)
    def _any_field_set(cls, values):
        """Validate that at least one field is set."""
        values_set = {k for k, v in values.items() if not isinstance(v, Unset)}
        if not values_set:
            raise ValueError("At least one field must be set.")
        return values


class HWMBulkCopyRequestV1(BaseModel):
    """Schema for request body of HWM copy operation."""

    source_namespace_id: int = Field(description="Source namespace ID from which HWMs are copied.")
    target_namespace_id: int = Field(description="Target namespace ID to which HWMs are copied.")
    hwm_ids: List[int] = Field(description="List of HWM IDs to be copied.")
    with_history: bool = Field(default=False, description="Whether to copy HWM history.")

    @validator("hwm_ids", pre=True, always=True)
    def _check_hwm_ids_not_empty(cls, v):
        if not len(v):
            raise ValueError("List should have at least 1 item after validation, not 0")
        return v

    @root_validator(skip_on_failure=True)
    def _check_namespace_ids(cls, values):
        """Validator to ensure source and target namespace IDs are different."""
        source_namespace_id, target_namespace_id = values.get("source_namespace_id"), values.get("target_namespace_id")
        if source_namespace_id == target_namespace_id:
            raise ValueError("Source and target namespace IDs must not be the same.")
        return values


class HWMBulkDeleteRequestV1(BaseModel):
    """Schema for request body of bulk delete HWM operation."""

    namespace_id: int
    hwm_ids: List[int]

    @validator("hwm_ids", pre=True, always=True)
    def _check_hwm_ids_not_empty(cls, v):
        if not len(v):
            raise ValueError("List should have at least 1 item after validation, not 0")
        return v



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

import logging
from time import time
from typing import Any, Dict, List, Optional, Tuple

from devtools import pformat
from fastapi import Depends, FastAPI
from typing_extensions import Annotated

from horizon.backend.db.models import User
from horizon.backend.dependencies import Stub
from horizon.backend.providers.auth.base import AuthProvider
from horizon.backend.services import UnitOfWork
from horizon.backend.settings.auth.dummy import DummyAuthProviderSettings
from horizon.backend.utils.jwt import decode_jwt, sign_jwt
from horizon.commons.exceptions.auth import AuthorizationError

log = logging.getLogger(__name__)


class DummyAuthProvider(AuthProvider):
    def __init__(
        self,
        settings: Annotated[DummyAuthProviderSettings, Depends(Stub(DummyAuthProviderSettings))],
        unit_of_work: Annotated[UnitOfWork, Depends()],
    ) -> None:
        self._settings = settings
        self._uow = unit_of_work

    @classmethod
    def setup(cls, app: FastAPI) -> FastAPI:
        settings = DummyAuthProviderSettings.parse_obj(app.state.settings.auth.dict(exclude={"provider"}))
        log.info("Using %s provider with settings:\n%s", cls.__name__, pformat(settings))
        app.dependency_overrides[AuthProvider] = cls
        app.dependency_overrides[DummyAuthProviderSettings] = lambda: settings
        return app

    async def get_current_user(self, access_token: str) -> User:
        if not access_token:
            raise AuthorizationError("Missing auth credentials")

        user_id = self._get_user_id_from_token(access_token)
        user = await self._uow.user.get_by_id(user_id)
        if not user.is_active:
            raise AuthorizationError(f"User {user.username!r} is disabled")
        return user

    async def get_token(
        self,
        grant_type: Optional[str] = None,
        login: Optional[str] = None,
        password: Optional[str] = None,
        scopes: Optional[List[str]] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ) -> Dict[str, Any]:
        if not login or not password:
            raise AuthorizationError("Missing auth credentials")

        log.info("Get/create user %r in database", login)
        async with self._uow:
            user = await self._uow.user.get_or_create(username=login)

        log.info("Used with id %r found", user.id)
        if not user.is_active:
            raise AuthorizationError(f"User {user.username!r} is disabled")

        log.info("Generate access token for user id %r", user.id)
        access_token, expires_at = self._generate_access_token(user_id=user.id)
        return {
            "access_token": access_token,
            "token_type": "bearer",
            "expires_at": expires_at,
        }

    def _generate_access_token(self, user_id: int) -> Tuple[str, float]:
        expires_at = time() + self._settings.access_token.expire_seconds
        payload = {
            "user_id": user_id,
            "exp": expires_at,
        }
        access_token = sign_jwt(
            payload,
            self._settings.access_token.secret_key.get_secret_value(),
            self._settings.access_token.security_algorithm,
        )
        return access_token, expires_at

    def _get_user_id_from_token(self, token: str) -> int:
        try:
            payload = decode_jwt(
                token,
                self._settings.access_token.secret_key.get_secret_value(),
                self._settings.access_token.security_algorithm,
            )
            return int(payload["user_id"])
        except (KeyError, TypeError, ValueError) as e:
            raise AuthorizationError("Invalid token") from e



# SPDX-FileCopyrightText: 2023-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

from horizon.backend.db.models import CredentialsCache
from horizon.backend.db.repositories.base import Repository


class CredentialsCacheRepository(Repository[CredentialsCache]):
    async def get_by_login(self, login: str) -> Optional[CredentialsCache]:
        return await self._get(CredentialsCache.login == login)

    async def create_or_update(
        self,
        login: str,
        data: dict,
    ) -> CredentialsCache:
        result = await self._update([CredentialsCache.login == login], changes=data)
        if not result:
            result = await self._create(data={"login": login, **data})

        await self._session.flush()
        return result

    async def delete(self, id: int) -> None:
        await self._delete(id)
        await self._session.flush()
